# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Element-wise functions expected by the python array API.
Internally, all functions apply an element-wise `operation` on:
        -> a python int/float/complex and a LinearSpaceElement
        -> two LinearSpaceElement
        -> a single LinearSpaceElement

Args:
    operation (str): a string identifier to lookup the desired function in the LinearSpaceElement's namespace.
    x1 (int | float | complex | LinearSpaceElement : Left operand
    x2 (int | float | complex | LinearSpaceElement (Optional) : Right operand. Defaults to None.
    out (LinearSpaceElement, optional): Out LinearSpaceElement for inplace updates. Defaults to None.

Returns:
    LinearSpaceElement: result of the element-wise operation on the array wrapped inside the element of an ODL space. 

Notes:
    1) The output array is wrapped in a space of which type depends of the output array's. This is a change of behaviour compared to ODL < 0.8.2
    2) Although one could use it to perform an operation on array-specific backend only, there is no clean way to infer a LinearSpace from the output. As such, one of the two operands must be a LinearSpaceElement
"""

__all__ = (
    'abs',
    'acos',
    'acosh',
    'add',
    'asin',
    'asinh',
    'atan',
    'atan2',
    'atanh',
    'bitwise_and',
    'bitwise_left_shift',
    'bitwise_invert',
    'bitwise_or',
    'bitwise_right_shift',
    'bitwise_xor',
    'ceil',
    'clip',
    'conj',
    'copysign',
    'cos',
    'cosh',
    'divide',
    'equal',
    'exp',
    'expm1',
    'floor',
    'floor_divide',
    'greater',
    'greater_equal',
    'hypot',
    'imag',
    'isfinite',
    'isinf',
    'isnan',
    'less',
    'less_equal',
    'log',
    'log1p',
    'log2',
    'log10',
    'logaddexp',
    'logical_and',
    'logical_not',
    'logical_or',
    'logical_xor',
    'maximum',
    'minimum',
    'multiply',
    'negative',
    'nextafter',
    'not_equal',
    'positive',
    'pow',
    'real',
    'reciprocal',
    'remainder',
    'round',
    'sign',
    'signbit',
    'sin',
    'sinh',
    'sqrt',
    'square',
    'subtract',
    'tan',
    'tanh',
    'trunc',
)


def _apply_element_wise(operation: str, x1, x2=None, out=None, **kwargs):
    """
    Examples
    >>> e0 = odl.rn(3).zero()
    >>> e1 = odl.rn(3).one()
    >>> e2 = e0 + e1
    >>> print(e2)
    [ 1.,  1.,  1.]
    >>> e3 = odl.add(e0, e1)
    >>> print(e3)
    [ 1.,  1.,  1.]
    >>> e2 == e3 
    True
    >>> e2 in odl.rn(3)
    True
    >>> new_el = e0 + 3j 
    >>> new_el in odl.rn(3)
    False
    >>> odl.add(odl.zeros_like(e1), e1)
    Traceback (most recent call last):
    TypeError: The type of the left operand <class 'numpy.ndarray'> is not supported.
    """
    # Lazy import of LinearSpaceElement and Operator for dispatching call
    from odl.core.operator import Operator
    from odl.core.set.space import LinearSpaceElement
    assert not isinstance(x1, Operator) or not isinstance(x2, Operator), f"ODL's array-API support for element-wise functions does not allow ODL Operators"
    if isinstance(x1, LinearSpaceElement):
        return x1.space._elementwise_num_operation(operation=operation, x1=x1, x2=x2, out=out, **kwargs)
    # Handling the left argument as a float/int/complex and right argument as a LinearSpaceElement
    elif isinstance(x2, LinearSpaceElement):
        return x2.space._elementwise_num_operation(operation=operation, x1=x1, x2=x2, out=out, **kwargs)   
    else:
        raise(AttributeError(f"Either x1 or x2 (if provided) need to be a LinearSpaceElement, got {type(x1)} and {type(x2)} with values {x1=} and {x2=}"))


def abs(x, out=None):
    """Calculates the absolute value for each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('abs', x, out=out)


def acos(x, out=None):
    """Calculates an implementation-dependent approximation of the principal
    value of the inverse cosine for each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('acos', x, out=out)


def acosh(x, out=None):
    """Calculates an implementation-dependent approximation to the inverse
    hyperbolic cosine for each element `x_i` of the input array `x`."""
    return _apply_element_wise('acosh', x, out=out)


def add(x1, x2, out=None):
    """Calculates the sum for each element `x1_i` of the input array `x1` with
    the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('add', x1, x2=x2, out=out)


def asin(x, out=None):
    """Calculates an implementation-dependent approximation of the principal
    value of the inverse sine for each element `x_i` of the input array `x`."""
    return _apply_element_wise('asin', x, out=out)


def asinh(x, out=None):
    """Calculates an implementation-dependent approximation to the inverse
    hyperbolic sine for each element `x_i` in the input array `x`."""
    return _apply_element_wise('asinh', x, out=out)


def atan(x, out=None):
    """Calculates an implementation-dependent approximation of the principal
    value of the inverse tangent for each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('atan', x, out=out)


def atan2(x1, x2, out=None):
    """Calculates an implementation-dependent approximation of the inverse
    tangent of the quotient `x1/x2`, having domain `[-infinity, +infinity]
    \times [-infinity, +infinity]` (where the `\times` notation denotes the set
    of ordered pairs of elements `(x1_i, x2_i)`) and codomain `[-pi, +pi]`,
    for each pair of elements `(x1_i, x2_i)` of the input arrays `x1` and `x2`,
    respectively."""
    return _apply_element_wise("atan2", x1, x2=x2, out=out)


def atanh(x, out=None):
    """Calculates an implementation-dependent approximation to the inverse
    hyperbolic tangent for each element `x_i` of the input array `x`."""
    return _apply_element_wise('atanh', x, out=out)


def bitwise_and(x1, x2, out=None):
    """Computes the bitwise AND of the underlying binary representation of each
    element `x1_i` of the input array `x1` with the respective element `x2_i`
    of the input array `x2`."""
    return _apply_element_wise('bitwise_and', x1, x2=x2, out=out)


def bitwise_left_shift(x1, x2, out=None):
    """Shifts the bits of each element `x1_i` of the input array `x1` to the
    left by appending `x2_i` (i.e., the respective element in the input array
    `x2`) zeros to the right of `x1_i`."""
    return _apply_element_wise('bitwise_left_shift', x1, x2=x2, out=out)


def bitwise_invert(x, out=None):
    """Inverts (flips) each bit for each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('bitwise_invert', x, out=out)


def bitwise_or(x1, x2, out=None):
    """Computes the bitwise OR of the underlying binary representation of each
    element `x1_i` of the input array `x1` with the respective element `x2_i`
    of the input array `x2`."""
    return _apply_element_wise('bitwise_or', x1, x2=x2, out=out)


def bitwise_right_shift(x1, x2, out=None):
    """Shifts the bits of each element `x1_i` of the input array `x1` to the
    right according to the respective element `x2_i` of the input array
    `x2`."""
    return _apply_element_wise('bitwise_right_shift', x1, x2=x2, out=out)


def bitwise_xor(x1, x2, out=None):
    """Computes the bitwise XOR of the underlying binary representation of each
    element `x1_i` of the input array `x1` with the respective element `x2_i`
    of the input array `x2`."""
    return _apply_element_wise('bitwise_xor', x1, x2=x2, out=out)


def ceil(x, out=None):
    """Rounds each element `x_i` of the input array `x` to the smallest (i.e.,
    closest to `-infty`) integer-valued number that is not less than `x_i`."""
    return _apply_element_wise('ceil', x, out=out)


def clip(x, out=None, min=None, max=None):
    """Clamps each element `x_i` of the input array `x` to the range `[min,
    max]`."""
    return _apply_element_wise('clip', x, out=out, min=min, max=max)


def conj(x, out=None):
    """Returns the complex conjugate for each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('conj', x, out=out)


def copysign(x1, x2, out=None):
    """Composes a floating-point value with the magnitude of `x1_i` and the
    sign of `x2_i` for each element of the input array `x1`."""
    return _apply_element_wise('copysign', x1, x2=x2, out=out)


def cos(x, out=None):
    """Calculates an implementation-dependent approximation to the cosine for
    each element `x_i` of the input array `x`."""
    return _apply_element_wise('cos', x, out=out)


def cosh(x, out=None):
    """Calculates an implementation-dependent approximation to the hyperbolic
    cosine for each element `x_i` in the input array `x`."""
    return _apply_element_wise('cosh', x, out=out)


def divide(x1, x2, out=None):
    """Calculates the division of each element `x1_i` of the input array `x1`
    with the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('divide', x1, x2=x2, out=out)


def equal(x1, x2, out=None):
    """Computes the truth value of `x1_i == x2_i` for each element `x1_i` of
    the input array `x1` with the respective element `x2_i` of the input array
    `x2`."""
    return _apply_element_wise('equal', x1, x2=x2, out=out)


def exp(x1, out=None):
    """Calculates an implementation-dependent approximation to the exponential
    function for each element `x_i` of the input array `x` (`e` raised to the
    power of `x_i`, where `e` is the base of the natural logarithm)."""
    return _apply_element_wise('exp', x1, out=out)


def expm1(x1, out=None):
    """Calculates an implementation-dependent approximation to `exp(x_i) - 1`
    for each element `x_i` of the input array `x`."""
    return _apply_element_wise("expm1", x1, out=out)


def floor(x1, out=None):
    """Rounds each element `x_i` of the input array `x` to the largest (i.e.,
    closest to `+infty`) integer-valued number that is not greater than
    `x_i`."""
    return _apply_element_wise('floor', x1, out=out)


def floor_divide(x1, x2, out=None):
    """Calculates the largest integer-valued number that is not greater than
    the result of dividing each element `x1_i` of the input array `x1` by the
    respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('floor_divide', x1, x2=x2, out=out)


def greater(x1, x2, out=None):
    """Computes the truth value of `x1_i > x2_i` for each element `x1_i` of the
    input array `x1` with the respective element `x2_i` of the input array
    `x2`."""
    return _apply_element_wise('greater', x1, x2=x2, out=out)


def greater_equal(x1, x2, out=None):
    """Computes the truth value of `x1_i >= x2_i` for each element `x1_i` of
    the input array `x1` with the respective element `x2_i` of the input array
    `x2`."""
    return _apply_element_wise('greater_equal', x1, x2=x2, out=out)


def hypot(x1, x2, out=None):
    """Computes the square root of the sum of squares for each element `x1_i`
    of the input array `x1` with the respective element `x2_i` of the input
    array `x2`."""
    return _apply_element_wise('hypot', x1, x2=x2, out=out)


def imag(x1, out=None):
    """Returns the imaginary part of each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('imag', x1, out=out)


def isfinite(x1, out=None):
    """Tests each element `x_i` of the input array `x` to determine if it is
    finite (i.e., not `NaN` and not an infinity)."""
    return _apply_element_wise('isfinite', x1, out=out)


def isinf(x1, out=None):
    """Tests each element `x_i` of the input array `x` to determine if it is a
    positive or negative infinity."""
    if x1 == float('Inf') or x1 == -float("Inf"):
        return True
    return _apply_element_wise('isinf', x1, out=out)


def isnan(x1, out=None):
    """Tests each element `x_i` of the input array `x` to determine if it is a
    `NaN`."""
    if x1 == float('Nan'):
        return True
    return _apply_element_wise('isnan', x1, out=out)


def less(x1, x2, out=None):
    """Computes the truth value of `x1_i < x2_i` for each element `x1_i` of the
    input array `x1` with the respective element `x2_i` of the input array
    `x2`."""
    return _apply_element_wise('less', x1, x2=x2, out=out)


def less_equal(x1, x2, out=None):
    """Computes the truth value of `x1_i <= x2_i` for each element `x1_i` of
    the input array `x1` with the respective element `x2_i` of the input array
    `x2`."""
    return _apply_element_wise('less_equal', x1, x2=x2, out=out)


def log(x1, out=None):
    """Calculates an implementation-dependent approximation to the natural
    logarithm for each element `x_i` of the input array `x`."""
    return _apply_element_wise('log', x1, out=out)


def log1p(x1, out=None):
    """Calculates an implementation-dependent approximation to `ln(1 + x_i)`
    for each element `x_i` of the input array `x`.

    For small `x`, the result of this function should be more accurate
    than `log(1 + x)`.
    """
    return _apply_element_wise("log1p", x1, out=out)


def log2(x1, out=None):
    """Calculates an implementation-dependent approximation to the base two
    logarithm for each element `x_i` of the input array `x`."""
    return _apply_element_wise("log2", x1, out=out)


def log10(x1, out=None):
    """Calculates an implementation-dependent approximation to the base ten
    logarithm for each element `x_i` of the input array `x`."""
    return _apply_element_wise("log10", x1, out=out)


def logaddexp(x1, x2, out=None):
    """Calculates the logarithm of the sum of exponentiations `log(exp(x1) +
    exp(x2))` for each element `x1_i` of the input array `x1` with the
    respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('logaddexp', x1, x2=x2, out=out)


def logical_and(x1, x2, out=None):
    """Computes the logical AND for each element `x1_i` of the input array `x1`
    with the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('logical_and', x1, x2=x2, out=out)


def logical_not(x1, out=None):
    """Computes the logical NOT for each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('logical_not', x1, out=out)


def logical_or(x1, x2, out=None):
    """Computes the logical OR for each element `x1_i` of the input array `x1`
    with the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('logical_or', x1, x2=x2, out=out)


def logical_xor(x1, x2, out=None):
    """Computes the logical XOR for each element `x1_i` of the input array `x1`
    with the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('logical_xor', x1, x2=x2, out=out)


def maximum(x1, x2, out=None):
    """Computes the maximum value for each element `x1_i` of the input array
    `x1` relative to the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('maximum', x1, x2=x2, out=out)


def minimum(x1, x2, out=None):
    """Calculates an implementation-dependent approximation of the principal
    value of the inverse cosine for each element."""
    return _apply_element_wise('minimum', x1, x2=x2, out=out)


def multiply(x1, x2, out=None):
    """Calculates the product for each element `x1_i` of the input array `x1`
    with the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('multiply', x1, x2=x2, out=out)


def negative(x1, out=None):
    """Numerically negates each element `x_i` of the input array `x`."""
    return _apply_element_wise('negative', x1, out=out)


def nextafter(x1, x2, out=None):
    """Returns the next representable floating-point value for each element
    `x1_i` of the input array `x1` in the direction of the respective element
    `x2_i` of the input array `x2`."""
    return _apply_element_wise('nextafter', x1, x2=x2, out=out)


def not_equal(x1, x2, out=None):
    """Computes the truth value of `x1_i != x2_i` for each element `x1_i` of
    the input array `x1` with the respective element `x2_i` of the input array
    `x2`."""
    return _apply_element_wise('not_equal', x1, x2=x2, out=out)


def positive(x1, out=None):
    """Numerically positive each element `x_i` of the input array `x`."""
    return _apply_element_wise('positive', x1, out=out)


def pow(x1, x2, out=None):
    """Calculates an implementation-dependent approximation of `x1_i` raised to
    the power of `x2_i` for each element `x1_i` of the input array `x1`, where
    `x2_i` is the corresponding element in the input array `x2`."""
    return _apply_element_wise('pow', x1, x2=x2, out=out)


def real(x1, out=None):
    """Returns the real part of each element `x_i` of the input array `x`."""
    return _apply_element_wise('real', x1, out=out)


def reciprocal(x1, out=None):
    """Returns the reciprocal for each element `x_i` of the input array `x`."""
    return _apply_element_wise('reciprocal', x1, out=out)


def remainder(x1, x2, out=None):
    """Calculates the remainder of dividing each element `x1_i` of the input
    array `x1` by the respective element `x2_i` of the input array `x2`.

    The result has the same sign as the dividend `x1`, and the magnitude
    is less than the magnitude of the divisor `x2`. This is often called
    the "Euclidean modulo" operation.
    """
    return _apply_element_wise('remainder', x1, x2=x2, out=out)


def round(x1, out=None):
    """Rounds each element `x_i` of the input array `x` to the nearest integer.

    Halfway cases (i.e., numbers with a fractional part of `0.5`) are rounded
    to the nearest even integer.
    """
    return _apply_element_wise('round', x1, out=out)


def sign(x1, out=None):
    """Returns an indication of the sign of each element `x_i` of the input
    array `x`.

    The returned array has the same shape as `x`.
    """
    return _apply_element_wise('sign', x1, out=out)


def signbit(x1, out=None):
    """Determines whether the sign bit is set for each element `x_i` of the
    input array `x`"""
    return _apply_element_wise('signbit', x1, out=out)


def sin(x1, out=None):
    """Calculates an implementation-dependent approximation to the sine for
    each element `x_i` of the input array `x`."""
    return _apply_element_wise('sin', x1, out=out)


def sinh(x1, out=None):
    """Calculates an implementation-dependent approximation to the hyperbolic
    sine for each element `x_i` in the input array `x`."""
    return _apply_element_wise('sinh', x1, out=out)


def sqrt(x1, out=None):
    """Calculates the square root for each element `x_i` of the input array
    `x`."""
    return _apply_element_wise('sqrt', x1, out=out)


def square(x1, out=None):
    """Calculates the square of each element `x_i` (i.e., `x_i * x_i`) of the
    input array `x`"""
    return _apply_element_wise('square', x1, out=out)


def subtract(x1, x2, out=None):
    """Calculates the difference for each element `x1_i` of the input array
    `x1` with the respective element `x2_i` of the input array `x2`."""
    return _apply_element_wise('subtract', x1, x2=x2, out=out)


def tan(x1, out=None):
    """Calculates an implementation-dependent approximation to the tangent for
    each element `x_i` of the input array `x`."""
    return _apply_element_wise('tan', x1, out=out)


def tanh(x1, out=None):
    """Calculates an implementation-dependent approximation to the hyperbolic
    tangent for each element `x_i` in the input array `x`."""
    return _apply_element_wise('tanh', x1, out=out)


def trunc(x1, out=None):
    """Rounds each element `x_i` of the input array `x` to the nearest integer
    towards zero."""
    return _apply_element_wise('trunc', x1, out=out)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for the factory functions to create proximal operators."""

from __future__ import division
import numpy as np
import scipy.special

import odl
from odl.solvers.nonsmooth.proximal_operators import (
    combine_proximals, proximal_const_func,
    proximal_box_constraint, proximal_nonnegativity,
    proximal_convex_conj_l1, proximal_convex_conj_l1_l2,
    proximal_l2,
    proximal_convex_conj_l2_squared,
    proximal_convex_conj_kl, proximal_convex_conj_kl_cross_entropy)
from odl.core.util.testutils import all_almost_equal
from odl.core.util.scipy_compatibility import lambertw

# Places for the accepted error when comparing results
HIGH_ACC = 8
LOW_ACC = 4


def test_proximal_const_func():
    """Proximal factory for the constnat mapping G(x) = c."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Element in the image space where the proximal operator is evaluated
    x = space.element(np.arange(-5, 5))

    # Factory function returning the proximal operator
    prox_factory = proximal_const_func(space)

    # Initialize proximal operator of G (with an unused parameter)
    prox = prox_factory(None)

    # prox_tau[G](x) = x = identity operator
    assert isinstance(prox, odl.IdentityOperator)

    # Optimal point of the auxiliary minimization problem prox_tau[G]
    x_opt = prox(x)

    # Identity map
    assert x == x_opt


def test_proximal_box_constraint():
    """Proximal factory for indicator function for non-negativity."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Element in the image space where the proximal operator is evaluated
    x = space.element(np.arange(-5, 5))

    for lower in [None, -2, -2.0 * space.one()]:
        for upper in [None, 2, 2.0 * space.one()]:
            # Factory function returning the proximal operator
            prox_factory = proximal_box_constraint(space,
                                                   lower=lower, upper=upper)
            prox = prox_factory(1.0)
            result = prox(x).asarray()

            # Create reference
            lower_np = -np.inf if lower is None else lower
            upper_np = np.inf if upper is None else upper
            result_np = odl.minimum(odl.maximum(x, lower_np), upper_np).asarray()

            # Verify equal result
            assert all_almost_equal(result_np, result)


def test_proximal_nonnegativity():
    """Proximal factory for indicator function for non-negativity."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Element in the image space where the proximal operator is evaluated
    x = space.element(np.arange(-5, 5))

    # Factory function returning the proximal operator
    prox_factory = proximal_nonnegativity(space)

    # Initialize proximal operator of G (with an unused parameter)
    prox = prox_factory(1.0)

    # Optimal point returned by the proximal operator
    result = prox(x)

    # prox_tau[G](x) = non-negativity thresholding
    assert all(result.asarray() >= 0)


def test_combine_proximal():
    """Function to combine proximal factory functions.

    The combine function makes use of the separable sum property of proximal
    operators.
    """

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Factory function returning the proximal operator
    prox_factory = proximal_const_func(space)

    # Combine factory function of proximal operators
    combined_prox_factory = combine_proximals(prox_factory, prox_factory)

    # Initialize combine proximal operator
    prox = combined_prox_factory(1)

    assert isinstance(prox, odl.Operator)

    # Explicit construction of the combine proximal operator
    prox_verify = odl.ProductSpaceOperator(
        [[odl.IdentityOperator(space), None],
         [None, odl.IdentityOperator(space)]])

    # Create an element in the domain of the operator
    x = prox_verify.domain.element([np.arange(-5, 5), np.arange(-5, 5)])

    # Allocate output element
    out = prox_verify.range.element()

    # Apply explicitly constructed and factory-function-combined proximal
    # operators
    assert prox(x) == prox_verify(x)

    # Test output argument
    assert prox(x, out) == prox_verify(x)

    # Identity mapping
    assert out == x


def test_proximal_l2_wo_data():
    """Proximal factory for the L2-norm."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Factory function returning the proximal operator
    lam = 2.0
    prox_factory = proximal_l2(space, lam=lam)

    # Initialize the proximal operator
    sigma = 3.0
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Elements
    x = space.element(np.arange(-5, 5))
    x_small = x * 0.5 * lam * sigma / x.norm()
    x_big = x * 2.0 * lam * sigma / x.norm()

    # Explicit computation:
    x_small_opt = x_small * 0
    x_big_opt = (1 - lam * sigma / x_big.norm()) * x_big

    assert all_almost_equal(prox(x_small), x_small_opt, HIGH_ACC)
    assert all_almost_equal(prox(x_big), x_big_opt, HIGH_ACC)


def test_proximal_l2_with_data():
    """Proximal factory for the L2-norm with data term."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Create data
    g = space.element(np.arange(-5, 5))

    # Factory function returning the proximal operator
    lam = 2.0
    prox_factory = proximal_l2(space, lam=lam, g=g)

    # Initialize the proximal operator
    sigma = 3.0
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Elements
    x = space.element(np.arange(-5, 5))
    x_small = g + x * 0.5 * lam * sigma / x.norm()
    x_big = g + x * 2.0 * lam * sigma / x.norm()

    # Explicit computation:
    x_small_opt = g
    const = lam * sigma / (x_big - g).norm()
    x_big_opt = (1 - const) * x_big + const * g

    assert all_almost_equal(prox(x_small), x_small_opt, HIGH_ACC)
    assert all_almost_equal(prox(x_big), x_big_opt, HIGH_ACC)


def test_proximal_convconj_l2_sq_wo_data():
    """Proximal factory for the convex conjugate of the L2-norm."""

    # Image space
    space = odl.uniform_discr(0, 10, 10)

    # Create an element in the image space
    x_arr = np.arange(-5, 5)
    x = space.element(x_arr)

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_l2_squared(space, lam=lam)

    # Initialize the proximal operators
    sigma = 0.25 * space.one()
    sigmav = sigma * space.one()
    prox = prox_factory(sigma)
    proxv = prox_factory(sigmav)

    assert isinstance(prox, odl.Operator)
    assert isinstance(proxv, odl.Operator)

    # Allocate output elements
    x_out = space.element()
    x_outv = space.element()

    # Optimal point returned by the proximal operator
    prox(x, x_out)
    proxv(x, x_outv)

    # Explicit computation: x / (1 + sigma / (2 * lambda))
    x_verify = x / (1 + sigma / (2 * lam))

    assert all_almost_equal(x_out, x_verify, HIGH_ACC)
    assert all_almost_equal(x_outv, x_verify, HIGH_ACC)


def test_proximal_convconj_l2_sq_with_data():
    """Proximal factory for the convex conjugate of the L2-norm."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Create an element in the image space
    x_arr = np.arange(-5, 5)
    x = space.element(x_arr)

    # Create data
    g = space.element(-2 * x_arr)

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_l2_squared(space, lam=lam, g=g)

    # Initialize the proximal operator
    sigma = 0.25
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Allocate output element
    x_out = space.element()

    # Optimal point returned by the proximal operator
    prox(x, x_out)

    # Explicit computation: (x - sigma * g) / (1 + sigma / (2 * lambda))
    x_verify = (x - sigma * g) / (1 + sigma / (2 * lam))

    assert all_almost_equal(x_out, x_verify, HIGH_ACC)


def test_proximal_convconj_l1_simple_space_without_data():
    """Proximal factory for the convex conjugate of the L1-norm."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Image element
    x_arr = np.arange(-5, 5)
    x = space.element(x_arr)

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_l1(space, lam=lam)

    # Initialize the proximal operator of F^*
    sigma = 0.25
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Apply the proximal operator returning its optimal point
    # Explicit computation: x / max(lam, |x|)
    denom = np.maximum(lam, np.sqrt(x_arr ** 2))
    x_exact = lam * x_arr / denom

    # Using out
    x_opt = space.element()
    x_result = prox(x, x_opt)
    assert x_result is x_opt
    assert all_almost_equal(x_opt, x_exact, HIGH_ACC)

    # Without out
    x_result = prox(x)
    assert all_almost_equal(x_result, x_exact, HIGH_ACC)

    # With aliased out
    x_result = prox(x, x)
    assert all_almost_equal(x_result, x_exact, HIGH_ACC)


def test_proximal_convconj_l1_simple_space_with_data():
    """Proximal factory for the convex conjugate of the L1-norm."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)
    x_arr = np.arange(-5, 5)
    x = space.element(x_arr)

    # RHS data
    g_arr = np.arange(10, 0, -1)
    g = space.element(g_arr)

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_l1(space, lam=lam, g=g)

    # Initialize the proximal operator of F^*
    sigma = 0.25
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Apply the proximal operator returning its optimal point
    x_opt = space.element()
    prox(x, x_opt)

    # Explicit computation: (x - sigma * g) / max(lam, |x - sigma * g|)
    denom = np.maximum(lam, np.abs(x_arr - sigma * g_arr))
    x0_verify = lam * (x_arr - sigma * g_arr) / denom

    assert all_almost_equal(x_opt, x0_verify, HIGH_ACC)


def test_proximal_convconj_l1_product_space():
    """Proximal factory for the convex conjugate of the L1-norm using
    product spaces."""

    # Product space for matrix of operators
    op_domain = odl.ProductSpace(odl.uniform_discr(0, 1, 10), 2)

    # Element in the product space where the proximal operator is evaluated
    x0_arr = np.arange(-5, 5)
    x1_arr = np.arange(10, 0, -1)
    x = op_domain.element([x0_arr, x1_arr])

    # Create a data element in the product space
    g0_arr = x1_arr.copy()
    g1_arr = x0_arr.copy()
    g = op_domain.element([g0_arr, g1_arr])

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_l1_l2(op_domain, lam=lam, g=g)

    # Initialize the proximal operator
    sigma = 0.25
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Apply the proximal operator returning its optimal point
    x_opt = prox(x)

    # Explicit computation: (x - sigma * g) / max(lam, |x - sigma * g|)
    denom = np.maximum(lam,
                       np.sqrt((x0_arr - sigma * g0_arr) ** 2 +
                               (x1_arr - sigma * g1_arr) ** 2))
    print(f'{denom=}')
    print(len([denom]))
    x_verify = lam * (x - sigma * g) / op_domain.element([denom])

    # Compare components
    assert all_almost_equal(x_verify, x_opt)


def test_proximal_convconj_kl_simple_space():
    """Test for proximal factory for the convex conjugate of KL divergence."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Element in image space where the proximal operator is evaluated
    x = space.element(np.arange(-5, 5))

    # Data
    g = space.element(np.arange(10, 0, -1))

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_kl(space, lam=lam, g=g)

    # Initialize the proximal operator of F^*
    sigma = 0.25
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Allocate an output element
    x_opt = space.element()

    # Apply the proximal operator returning its optimal point
    prox(x, x_opt)

    # Explicit computation:
    x_verify = (lam + x - odl.sqrt((x - lam) ** 2 + 4 * lam * sigma * g)) / 2

    assert all_almost_equal(x_opt, x_verify, HIGH_ACC)


def test_proximal_convconj_kl_product_space():
    """Test for product spaces in proximal for conjugate of KL divergence"""

    # Product space for matrix of operators
    op_domain = odl.ProductSpace(odl.uniform_discr(0, 1, 10), 2)

    # Element in the product space where the proximal operator is evaluated
    x0_arr = np.arange(-5, 5)
    x1_arr = np.arange(10, 0, -1)
    x = op_domain.element([x0_arr, x1_arr])

    # Element in the product space with given data
    g0_arr = x1_arr.copy()
    g1_arr = x0_arr.copy()
    g = op_domain.element([g0_arr, g1_arr])

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_kl(op_domain, lam=lam, g=g)

    # Initialize the proximal operator
    sigma = 0.25
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Allocate an output element
    x_opt = op_domain.element()

    # Apply the proximal operator returning its optimal point
    prox(x, x_opt)

    # Explicit computation:
    x_verify = (lam + x - odl.sqrt((x - lam) ** 2 + 4 * lam * sigma * g)) / 2

    # Compare components
    assert all_almost_equal(x_verify, x_opt)


def test_proximal_convconj_kl_cross_entropy():
    """Test for proximal of convex conjugate of cross entropy KL divergence."""

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Data
    g = space.element(np.arange(10, 0, -1))

    # Factory function returning the proximal operator
    lam = 2
    prox_factory = proximal_convex_conj_kl_cross_entropy(space, lam=lam, g=g)

    # Initialize the proximal operator of F^*
    sigma = 0.25
    prox = prox_factory(sigma)

    assert isinstance(prox, odl.Operator)

    # Element in image space where the proximal operator is evaluated
    x = space.element(np.arange(-5, 5))

    prox_val = prox(x)

    # Explicit computation:
    x_verify = x - lam * lambertw(
        sigma / lam * g * odl.exp(x / lam)).real

    assert all_almost_equal(prox_val, x_verify, HIGH_ACC)

    # Test in-place evaluation
    x_inplace = space.element()
    prox(x, out=x_inplace)

    assert all_almost_equal(x_inplace, x_verify, HIGH_ACC)


def test_proximal_arg_scaling():
    """Test for proximal argument scaling."""

    # Set the underlying space.
    space = odl.uniform_discr(0, 1, 10)

    # Set the functional and the prox factory.
    func = odl.functionals.L2NormSquared(space)
    prox_factory = odl.solvers.proximal_l2_squared(space)

    # Set the point where the proximal operator will be evaluated.
    x = space.one()

    # Set the scaling parameters.
    for alpha in [2, odl.core.phantom.noise.uniform_noise(space, 1, 10)]:
        # Scale the proximal factories
        prox_scaled = odl.solvers.proximal_arg_scaling(prox_factory, alpha)

        # Set the step size.
        for sigma in [2, odl.core.phantom.noise.uniform_noise(space, 1, 10)]:
            # Evaluation of the proximals
            p = prox_scaled(sigma)(x)

            # Now we know that p = Prox_{sigma g}(x) where g(x) = f(alpha x),
            # i.e., (x - p)/sigma = grad g(p) = alpha * grad f(alpha p).
            lhs = (x - p) / sigma
            rhs = alpha * func.gradient(alpha * p)
            assert all_almost_equal(lhs, rhs)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Functions to create noise samples of different distributions."""

from __future__ import absolute_import, division, print_function

import numpy as np

from odl.core.util import npy_random_seed
from odl.core.space.base_tensors import Tensor

__all__ = ('white_noise', 'poisson_noise', 'salt_pepper_noise',
           'uniform_noise')


def white_noise(space, mean=0, stddev=1, seed=None):
    """Standard gaussian noise in space, pointwise ``N(mean, stddev**2)``.

    Parameters
    ----------
    space : `TensorSpace` or `ProductSpace`
        The space in which the noise is created.
    mean : ``space.field`` element or ``space`` `element-like`, optional
        The mean of the white noise. If a scalar, it is interpreted as
        ``mean * space.one()``.
        If ``space`` is complex, the real and imaginary parts are interpreted
        as the mean of their respective part of the noise.
    stddev : `float` or ``space`` `element-like`, optional
        The standard deviation of the white noise. If a scalar, it is
        interpreted as ``stddev * space.one()``.
    seed : int, optional
        Random seed to use for generating the noise.
        For ``None``, use the current seed.

    Returns
    -------
    white_noise : ``space`` element

    See Also
    --------
    poisson_noise
    salt_pepper_noise
    numpy.random.normal
    """
    from odl.core.space import ProductSpace

    with npy_random_seed(seed):
        if isinstance(space, ProductSpace):
            values = [white_noise(subspace, mean, stddev)
                      for subspace in space]
        else:
            if space.is_complex:
                real = np.random.normal(
                    loc=mean.real, scale=stddev, size=space.shape)
                imag = np.random.normal(
                    loc=mean.imag, scale=stddev, size=space.shape)
                values = real + 1j * imag
            else:
                if isinstance(mean, Tensor):
                    array_backend = mean.array_backend
                    mean = array_backend.to_numpy(mean)
                
                values = np.random.normal(
                    loc=mean, scale=stddev, size=space.shape)

    return space.element(values)


def uniform_noise(space, low=0, high=1, seed=None):
    """Uniformly distributed noise in ``space``, pointwise ``U(low, high)``.

    Parameters
    ----------
    space : `TensorSpace` or `ProductSpace`
        The space in which the noise is created.
    low : ``space.field`` element or ``space`` `element-like`, optional
        The lower bound of the uniform noise. If a scalar, it is interpreted as
        ``low * space.one()``.
        If ``space`` is complex, the real and imaginary parts are interpreted
        as their respective part of the noise.
    high : ``space.field`` element or ``space`` `element-like`, optional
        The upper bound of the uniform noise. If a scalar, it is interpreted as
        ``high * space.one()``.
        If ``space`` is complex, the real and imaginary parts are interpreted
        as their respective part of the noise.
    seed : int, optional
        Random seed to use for generating the noise.
        For ``None``, use the current seed.

    Returns
    -------
    white_noise : ``space`` element

    See Also
    --------
    poisson_noise
    salt_pepper_noise
    white_noise
    numpy.random.normal
    """
    from odl.core.space import ProductSpace

    with npy_random_seed(seed):
        if isinstance(space, ProductSpace):
            values = [uniform_noise(subspace, low, high)
                      for subspace in space]
        else:
            if space.is_complex:
                real = np.random.uniform(low=low.real, high=high.real,
                                         size=space.shape)
                imag = np.random.uniform(low=low.imag, high=high.imag,
                                         size=space.shape)
                values = real + 1j * imag
            else:
                values = np.random.uniform(low=low, high=high,
                                           size=space.shape)

    return space.element(values)


def poisson_noise(intensity, seed=None):
    r"""Poisson distributed noise with given intensity.

    Parameters
    ----------
    intensity : `TensorSpace` or `ProductSpace` element
        The intensity (usually called lambda) parameter of the noise.

    Returns
    -------
    poisson_noise : ``intensity.space`` element
        Poisson distributed random variable.
    seed : int, optional
        Random seed to use for generating the noise.
        For ``None``, use the current seed.

    Notes
    -----
    For a Poisson distributed random variable :math:`X` with intensity
    :math:`\lambda`, the probability of it taking the value
    :math:`k \in \mathbb{N}_0` is given by

    .. math::
        \frac{\lambda^k e^{-\lambda}}{k!}

    Note that the function only takes integer values.

    See Also
    --------
    white_noise
    salt_pepper_noise
    uniform_noise
    numpy.random.poisson
    """
    from odl.core.space import ProductSpace

    with npy_random_seed(seed):
        if isinstance(intensity.space, ProductSpace):
            values = [poisson_noise(subintensity)
                      for subintensity in intensity]
        else:
            values = np.random.poisson(intensity.asarray())

    return intensity.space.element(values)


def salt_pepper_noise(vector, fraction=0.05, salt_vs_pepper=0.5,
                      low_val=None, high_val=None, seed=None):
    """Add salt and pepper noise to vector.

    Salt and pepper noise replaces random elements in ``vector`` with
    ``low_val`` or ``high_val``.

    Parameters
    ----------
    vector : element of `TensorSpace` or `ProductSpace`
        The vector that noise should be added to.
    fraction : float, optional
        The propotion of the elements in ``vector`` that should be converted
        to noise.
    salt_vs_pepper : float, optional
        Relative abundance of salt (high) vs pepper (low) noise. A high value
        means more salt than pepper noise.
    low_val : float, optional
        The "pepper" color in the noise.
        Default: minimum value of ``vector``. For product spaces the minimum
        value per subspace is taken.
    high_val : float, optional
        The "salt" value in the noise.
        Default: maximuim value of ``vector``. For product spaces the maximum
        value per subspace is taken.
    seed : int, optional
        Random seed to use for generating the noise.
        For ``None``, use the current seed.

    Returns
    -------
    salt_pepper_noise : ``vector.space`` element
        ``vector`` with salt and pepper noise.

    See Also
    --------
    white_noise
    poisson_noise
    uniform_noise
    """
    from odl.core.space import ProductSpace

    # Validate input parameters
    fraction, fraction_in = float(fraction), fraction
    if not (0 <= fraction <= 1):
        raise ValueError('`fraction` ({}) should be a float in the interval '
                         '[0, 1]'.format(fraction_in))

    salt_vs_pepper, salt_vs_pepper_in = float(salt_vs_pepper), salt_vs_pepper
    if not (0 <= salt_vs_pepper <= 1):
        raise ValueError('`salt_vs_pepper` ({}) should be a float in the '
                         'interval [0, 1]'.format(salt_vs_pepper_in))

    with npy_random_seed(seed):
        if isinstance(vector.space, ProductSpace):
            values = [salt_pepper_noise(subintensity, fraction, salt_vs_pepper,
                                        low_val, high_val)
                      for subintensity in vector]
        else:
            # Extract vector of values
            values = vector.asarray().flatten()

            # Determine fill-in values if not given
            if low_val is None:
                low_val = np.min(values)
            if high_val is None:
                high_val = np.max(values)

            # Create randomly selected points as a subset of image.
            a = np.arange(vector.size)
            np.random.shuffle(a)
            salt_indices = a[:int(fraction * vector.size * salt_vs_pepper)]
            pepper_indices = a[int(fraction * vector.size * salt_vs_pepper):
                               int(fraction * vector.size)]

            values[salt_indices] = high_val
            values[pepper_indices] = low_val
            values = values.reshape(vector.space.shape)

    return vector.space.element(values)


if __name__ == '__main__':
    # Show the phantoms
    import odl
    from odl.core.util.testutils import run_doctests

    r100 = odl.rn(100)
    white_noise(r100).show('white_noise')
    uniform_noise(r100).show('uniform_noise')
    white_noise(r100, mean=5).show('white_noise with mean')

    c100 = odl.cn(100)
    white_noise(c100).show('complex white_noise')
    uniform_noise(c100).show('complex uniform_noise')

    discr = odl.uniform_discr([-1, -1], [1, 1], [300, 300])
    white_noise(discr).show('white_noise 2d')
    uniform_noise(discr).show('uniform_noise 2d')

    vector = odl.core.phantom.shepp_logan(discr, modified=True)
    poisson_noise(vector * 100).show('poisson_noise 2d')
    salt_pepper_noise(vector).show('salt_pepper_noise 2d')

    # Run also the doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the smooth solvers."""

from __future__ import division
import pytest
import odl
from odl.core.operator import OpNotImplementedError


nonlinear_cg_beta = odl.core.util.testutils.simple_fixture('nonlinear_cg_beta',
                                                      ['FR', 'PR', 'HS', 'DY'])


@pytest.fixture(scope="module", params=['l2_squared', 'l2_squared_scaled',
                                        'rosenbrock', 'quadratic_form'])
def functional(request):
    """functional with optimum 0 at 0."""
    name = request.param

    if name == 'l2_squared':
        space = odl.rn(3)
        return odl.functionals.L2NormSquared(space)
    elif name == 'l2_squared_scaled':
        space = odl.uniform_discr(0, 1, 3)
        scaling = odl.MultiplyOperator(space.element([1, 2, 3]),
                                       domain=space)
        return odl.functionals.L2NormSquared(space) * scaling
    elif name == 'quadratic_form':
        space = odl.rn(3)
        # Symmetric and diagonally dominant matrix
        matrix = odl.MatrixOperator([[7.0, 1, 2],
                                     [1, 5, -3],
                                     [2, -3, 8]])
        vector = space.element([1, 2, 3])

        # Calibrate so that functional is zero in optimal point
        constant = 1 / 4 * vector.inner(matrix.inverse(vector))

        return odl.functionals.QuadraticForm(
            operator=matrix, vector=vector, constant=constant)
    elif name == 'rosenbrock':
        # Moderately ill-behaved rosenbrock functional.
        rosenbrock = odl.functionals.RosenbrockFunctional(odl.rn(2), scale=2)

        # Center at zero
        return rosenbrock.translated([-1, -1])
    else:
        assert False


@pytest.fixture(scope="module", params=['constant', 'backtracking'])
def functional_and_linesearch(request, functional):
    """Return functional with optimum 0 at 0 and a line search."""
    name = request.param

    if name == 'constant':
        return functional, 1.0
    else:
        return functional, odl.solvers.BacktrackingLineSearch(functional)


@pytest.fixture(scope="module", params=['first', 'second'])
def broyden_impl(request):
    return request.param


def test_newton_solver(functional_and_linesearch):
    """Test the Newton solver."""
    functional, line_search = functional_and_linesearch

    try:
        # Test if derivative exists
        functional.gradient.derivative(functional.domain.zero())
    except OpNotImplementedError:
        return

    # Solving the problem
    x = functional.domain.one()
    odl.solvers.newtons_method(functional, x, tol=1e-6,
                               line_search=line_search)

    # Assert x is close to the optimum at [1, 1]
    assert functional(x) < 1e-3


def test_bfgs_solver(functional_and_linesearch):
    """Test the BFGS quasi-Newton solver."""
    functional, line_search = functional_and_linesearch

    x = functional.domain.one()
    odl.solvers.bfgs_method(functional, x, tol=1e-3,
                            line_search=line_search)

    assert functional(x) < 1e-3


def test_lbfgs_solver(functional_and_linesearch):
    """Test limited memory BFGS quasi-Newton solver."""
    functional, line_search = functional_and_linesearch

    x = functional.domain.one()
    odl.solvers.bfgs_method(functional, x, tol=1e-3,
                            line_search=line_search, num_store=5)

    assert functional(x) < 1e-3


def test_broydens_method(broyden_impl, functional_and_linesearch):
    """Test the ``broydens_method`` quasi-Newton solver."""
    functional, line_search = functional_and_linesearch

    x = functional.domain.one()
    odl.solvers.broydens_method(functional, x, tol=1e-3,
                                line_search=line_search, impl=broyden_impl)

    assert functional(x) < 1e-3


def test_steepest_descent(functional):
    """Test the ``steepest_descent`` solver."""
    line_search = odl.solvers.BacktrackingLineSearch(functional)

    x = functional.domain.one()
    odl.solvers.steepest_descent(functional, x, tol=1e-3,
                                 line_search=line_search)

    assert functional(x) < 1e-3


def test_adam(functional):
    """Test the ``adam`` solver."""

    x = functional.domain.one()
    odl.solvers.adam(functional, x, tol=1e-2, learning_rate=0.5)

    assert functional(x) < 1e-3


def test_conjguate_gradient_nonlinear(functional, nonlinear_cg_beta):
    """Test the ``conjugate_gradient_nonlinear`` solver."""
    line_search = odl.solvers.BacktrackingLineSearch(functional)

    x = functional.domain.one()
    odl.solvers.conjugate_gradient_nonlinear(functional, x, tol=1e-3,
                                             line_search=line_search,
                                             beta_method=nonlinear_cg_beta)

    assert functional(x) < 1e-3


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for the Ray transform."""

from __future__ import division

import numpy as np
import math
import pytest
from packaging.version import parse as parse_version
from functools import partial

import odl
from odl.applications.tomo.backends import ASTRA_AVAILABLE, ASTRA_VERSION
from odl.applications.tomo.util.testutils import (
    skip_if_no_astra, skip_if_no_astra_cuda, skip_if_no_skimage, skip_if_no_pytorch)
from odl.core.util.testutils import all_equal, all_almost_equal, simple_fixture
from odl.core.array_API_support.comparisons  import odl_all_equal
# --- pytest fixtures --- #


impl = simple_fixture(
    name='impl',
    params=[pytest.param('astra_cpu', marks=skip_if_no_astra),
            pytest.param('astra_cuda', marks=skip_if_no_astra_cuda),
            pytest.param('skimage', marks=skip_if_no_skimage)]
)

geometry_params = ['par2d', 'par3d', 'cone2d', 'cone3d', 'helical']
geometry_ids = [" geometry='{}' ".format(p) for p in geometry_params]


@pytest.fixture(scope='module', ids=geometry_ids, params=geometry_params)
def geometry(request):
    geom = request.param
    m = 100
    n_angles = 100

    if geom == 'par2d':
        apart = odl.uniform_partition(0, np.pi, n_angles)
        dpart = odl.uniform_partition(-30, 30, m)
        return odl.applications.tomo.Parallel2dGeometry(apart, dpart)
    elif geom == 'par3d':
        apart = odl.uniform_partition(0, np.pi, n_angles)
        dpart = odl.uniform_partition([-30, -30], [30, 30], (m, m))
        return odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart)
    elif geom == 'cone2d':
        apart = odl.uniform_partition(0, 2 * np.pi, n_angles)
        dpart = odl.uniform_partition(-30, 30, m)
        return odl.applications.tomo.FanBeamGeometry(apart, dpart, src_radius=200,
                                        det_radius=100)
    elif geom == 'cone3d':
        apart = odl.uniform_partition(0, 2 * np.pi, n_angles)
        dpart = odl.uniform_partition([-60, -60], [60, 60], (m, m))
        return odl.applications.tomo.ConeBeamGeometry(apart, dpart,
                                         src_radius=200, det_radius=100)
    elif geom == 'helical':
        apart = odl.uniform_partition(0, 8 * 2 * np.pi, n_angles)
        dpart = odl.uniform_partition([-30, -3], [30, 3], (m, m))
        return odl.applications.tomo.ConeBeamGeometry(apart, dpart, pitch=5.0,
                                         src_radius=200, det_radius=100)
    else:
        raise ValueError('geom not valid')


geometry_type = simple_fixture(
    'geometry_type',
    ['par2d', 'par3d', 'cone2d', 'cone3d']
)

projectors = []
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_astra)
     for proj_cfg in ['par2d astra_cpu uniform numpy cpu',
                      'par2d astra_cpu nonuniform numpy cpu',
                      'par2d astra_cpu random numpy cpu',
                      'cone2d astra_cpu uniform numpy cpu',
                      'cone2d astra_cpu nonuniform numpy cpu',
                      'cone2d astra_cpu random numpy cpu',
                      ])
)
projectors.extend(
    (pytest.param(proj_cfg, marks=[skip_if_no_astra, skip_if_no_pytorch])
     for proj_cfg in [
                      'par2d astra_cpu uniform pytorch cpu',
                      'par2d astra_cpu nonuniform pytorch cpu',
                      'par2d astra_cpu random pytorch cpu',
                      'cone2d astra_cpu uniform pytorch cpu',
                      'cone2d astra_cpu nonuniform pytorch cpu',
                      'cone2d astra_cpu random pytorch cpu'
                      ])

)
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_astra_cuda)
     for proj_cfg in ['par2d astra_cuda uniform numpy cpu',
                      'par2d astra_cuda half_uniform numpy cpu',
                      'par2d astra_cuda nonuniform numpy cpu',
                      'par2d astra_cuda random numpy cpu',
                      'cone2d astra_cuda uniform numpy cpu',
                      'cone2d astra_cuda nonuniform numpy cpu',
                      'cone2d astra_cuda random numpy cpu',
                      'par3d astra_cuda uniform numpy cpu',
                      'par3d astra_cuda nonuniform numpy cpu',
                      'par3d astra_cuda random numpy cpu',
                      'cone3d astra_cuda uniform numpy cpu',
                      'cone3d astra_cuda nonuniform numpy cpu',
                      'cone3d astra_cuda random numpy cpu',
                      ])
)

projectors.extend(
    (pytest.param(proj_cfg, marks=[skip_if_no_astra, skip_if_no_pytorch])
     for proj_cfg in [
                      'par2d astra_cuda uniform pytorch cuda:0',
                      'par2d astra_cuda half_uniform pytorch cuda:0',
                      'par2d astra_cuda nonuniform pytorch cuda:0',
                      'par2d astra_cuda random pytorch cuda:0',
                      'cone2d astra_cuda uniform pytorch cuda:0',
                      'cone2d astra_cuda nonuniform pytorch cuda:0',
                      'cone2d astra_cuda random pytorch cuda:0',
                      'par3d astra_cuda uniform pytorch cuda:0',
                      'par3d astra_cuda nonuniform pytorch cuda:0',
                      'par3d astra_cuda random pytorch cuda:0',
                      'cone3d astra_cuda uniform pytorch cuda:0',
                      'cone3d astra_cuda nonuniform pytorch cuda:0',
                      'cone3d astra_cuda random pytorch cuda:0',
                      'helical astra_cuda uniform pytorch cuda:0'])
)

projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_skimage)
     for proj_cfg in ['par2d skimage uniform numpy cpu',
                      'par2d skimage half_uniform numpy cpu'])
)

projector_ids = [
    " geom='{}' - astra_impl='{}' - angles='{}' - tspace_impl='{}' - tspace_device='{}'".format(*p.values[0].split())
    for p in projectors
]

@pytest.fixture(scope='module', params=projectors, ids=projector_ids)
def projector(request):
    n = 100
    m = 100
    n_angles = 100
    dtype = 'float32'
    geom, astra_impl, angle, tspace_impl, tspace_device = request.param.split()

    if angle == 'uniform':
        apart = odl.uniform_partition(0, 2 * np.pi, n_angles)
    elif angle == 'half_uniform':
        apart = odl.uniform_partition(0, np.pi, n_angles)
    elif angle == 'random':
        # Linearly spaced with random noise
        min_pt = 2 * (2.0 * np.pi) / n_angles
        max_pt = (2.0 * np.pi) - 2 * (2.0 * np.pi) / n_angles
        points = np.linspace(min_pt, max_pt, n_angles)
        points += np.random.rand(n_angles) * (max_pt - min_pt) / (5 * n_angles)
        apart = odl.nonuniform_partition(points)
    elif angle == 'nonuniform':
        # Angles spaced quadratically
        min_pt = 2 * (2.0 * np.pi) / n_angles
        max_pt = (2.0 * np.pi) - 2 * (2.0 * np.pi) / n_angles
        points = np.linspace(min_pt ** 0.5, max_pt ** 0.5, n_angles) ** 2
        apart = odl.nonuniform_partition(points)
    else:
        raise ValueError('angle not valid')

    if geom == 'par2d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 2, [20] * 2, [n] * 2,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry
        dpart = odl.uniform_partition(-30, 30, m)
        geom = odl.applications.tomo.Parallel2dGeometry(apart, dpart)

    elif geom == 'par3d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 3, [20] * 3, [n] * 3,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)

        # Geometry
        dpart = odl.uniform_partition([-30] * 2, [30] * 2, [m] * 2)
        geom = odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart)

    elif geom == 'cone2d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 2, [20] * 2, [n] * 2,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry
        dpart = odl.uniform_partition(-30, 30, m)
        geom = odl.applications.tomo.FanBeamGeometry(apart, dpart, src_radius=200,
                                        det_radius=100)

    elif geom == 'cone3d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 3, [20] * 3, [n] * 3,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry
        dpart = odl.uniform_partition([-60] * 2, [60] * 2, [m] * 2)
        geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart,
                                         src_radius=200, det_radius=100)

    elif geom == 'helical':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20, -20, 0], [20, 20, 40],
                                       [n] * 3, dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry, overwriting angle partition
        apart = odl.uniform_partition(0, 8 * 2 * np.pi, n_angles)
        dpart = odl.uniform_partition([-30, -3], [30, 3], [m] * 2)
        geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart, pitch=5.0,
                                         src_radius=200, det_radius=100)
    else:
        raise ValueError('geom not valid')
    
    # Ray transform
    return odl.applications.tomo.RayTransform(reco_space, geom, impl=astra_impl, use_cache=False)


@pytest.fixture(scope='module',
                params=[True, False],
                ids=[' in-place ', ' out-of-place '])
def in_place(request):
    return request.param


# --- RayTransform tests --- #


def test_projector(projector, in_place):
    """Test Ray transform forward projection."""
    # TODO: this needs to be improved
    # Accept 10% errors
    rtol = 1e-1

    # Create Shepp-Logan phantom
    vol = projector.domain.one()

    # Calculate projection
    if in_place:
        proj = projector.range.zero()
        projector(vol, out=proj)
    else:
        proj = projector(vol)

    # We expect maximum value to be along diagonal
    expected_max = projector.domain.partition.extent[0] * math.sqrt(2)
    assert odl.max(proj) == pytest.approx(expected_max, rel=rtol)


def test_adjoint(projector):
    """Test Ray transform backward projection."""
    # Relative tolerance, still rather high due to imperfectly matched
    # adjoint in the cone beam case
    if (
        ASTRA_AVAILABLE
        and parse_version(ASTRA_VERSION) < parse_version('1.8rc1')
        and isinstance(projector.geometry, odl.applications.tomo.ConeBeamGeometry)
    ):
        rtol = 0.1
    else:
        rtol = 0.05

    # Create Shepp-Logan phantom
    vol = odl.core.phantom.shepp_logan(projector.domain, modified=True)

    # Calculate projection
    proj = projector(vol)
    backproj = projector.adjoint(proj)

    # Verified the identity <Ax, Ax> = <A^* A x, x>
    result_AxAx = proj.inner(proj)
    result_xAtAx = backproj.inner(vol)
    assert result_AxAx == pytest.approx(result_xAtAx, rel=rtol)


def test_adjoint_of_adjoint(projector):
    """Test Ray transform adjoint of adjoint."""

    # Create Shepp-Logan phantom
    vol = odl.core.phantom.shepp_logan(projector.domain, modified=True)

    # Calculate projection
    proj = projector(vol)
    proj_adj_adj = projector.adjoint.adjoint(vol)

    # Verify A(x) == (A^*)^*(x)
    assert all_almost_equal(proj, proj_adj_adj)

    # Calculate adjoints
    proj_adj = projector.adjoint(proj)
    proj_adj_adj_adj = projector.adjoint.adjoint.adjoint(proj)

    # Verify A^*(y) == ((A^*)^*)^*(x)
    assert all_almost_equal(proj_adj, proj_adj_adj_adj)


def test_angles(projector):
    """Test Ray transform angle conventions."""

    ns = projector.domain.tspace.array_namespace

    # Smoothed line/hyperplane with offset
    vol = projector.domain.element(
        lambda x: ns.exp(-(2 * x[0] - 10 + x[1]) ** 2))

    # Create projection
    result = projector(vol).asarray()    

    # Find the angle where the projection has a maximum (along the line).
    # TODO: center of mass would be more robust
    axes = 1 if projector.domain.ndim == 2 else (1, 2)
    ind_angle = ns.argmax(ns.max(result, axis=axes))
    # Restrict to [0, 2 * pi) for helical
    maximum_angle = np.fmod(projector.geometry.angles[ind_angle], 2 * np.pi)

    # Verify correct maximum angle. The line is defined by the equation
    # x1 = 10 - 2 * x0, i.e. the slope -2. Thus the angle arctan(1/2) should
    # give the maximum projection values.
    expected = np.arctan2(1, 2)
    assert np.fmod(maximum_angle, np.pi) == pytest.approx(expected, abs=0.1)

    # Find the pixel where the projection has a maximum at that angle
    axes = () if projector.domain.ndim == 2 else 1
    ind_pixel = ns.argmax(ns.max(result[ind_angle], axis=axes))
    
    # We must convert the ind_pixel back to a float on the cpu
    if projector.domain.tspace.impl == 'pytorch':
        ind_pixel = int(ind_pixel.detach().cpu())

    max_pixel = projector.geometry.det_partition[ind_pixel, ...].mid_pt[0]

    # The line is at distance 2 * sqrt(5) from the origin, which translates
    # to the same distance from the detector midpoint, with positive sign
    # if the angle is smaller than pi and negative sign otherwise.
    expected = 2 * math.sqrt(5) if maximum_angle < np.pi else -2 * math.sqrt(5)

    # We need to scale with the magnification factor if applicable
    if isinstance(projector.geometry, odl.applications.tomo.DivergentBeamGeometry):
        src_to_det = (
            projector.geometry.src_radius
            + projector.geometry.det_radius
        )
        magnification = src_to_det / projector.geometry.src_radius
        expected *= magnification

    assert max_pixel == pytest.approx(expected, abs=0.2)


def test_complex(impl, odl_impl_device_pairs):
    tspace_impl, device = odl_impl_device_pairs
    """Test transform of complex input for parallel 2d geometry."""
    space_c = odl.uniform_discr([-1, -1], [1, 1], (10, 10), dtype='complex64', impl=tspace_impl, device=device)
    space_r = space_c.real_space
    geom = odl.applications.tomo.parallel_beam_geometry(space_c)

    if tspace_impl == 'pytorch' and impl == 'skimage':
        pytest.skip(f'Skimage backend not available with pytorch') 
    
    ray_trafo_c = odl.applications.tomo.RayTransform(space_c, geom, impl=impl)
    ray_trafo_r = odl.applications.tomo.RayTransform(space_r, geom, impl=impl)
    vol = odl.core.phantom.shepp_logan(space_c)
    vol.imag = odl.core.phantom.cuboid(space_r)

    data = ray_trafo_c(vol)
    true_data_re = ray_trafo_r(vol.real)
    true_data_im = ray_trafo_r(vol.imag)

    assert all_equal(data.real, true_data_re)
    assert all_equal(data.imag, true_data_im)

    # test adjoint for complex data
    backproj_r = ray_trafo_r.adjoint
    backproj_c = ray_trafo_c.adjoint
    true_vol_re = backproj_r(data.real)
    true_vol_im = backproj_r(data.imag)
    backproj_vol = backproj_c(data)

    assert all_equal(backproj_vol.real, true_vol_re)
    assert all_equal(backproj_vol.imag, true_vol_im)


def test_anisotropic_voxels(geometry, odl_impl_device_pairs):
    tspace_impl, device = odl_impl_device_pairs
    """Test projection and backprojection with anisotropic voxels."""
    ndim = geometry.ndim
    shape = [10] * (ndim - 1) + [5]
    space = odl.uniform_discr([-1] * ndim, [1] * ndim, shape=shape,
                              dtype='float32', impl=tspace_impl, device=device)

    # If no implementation is available, skip
    if ndim == 2 and not odl.applications.tomo.ASTRA_AVAILABLE:
        pytest.skip(reason='ASTRA not available, skipping 2d test')
    elif ndim == 3 and not odl.applications.tomo.ASTRA_CUDA_AVAILABLE:
        pytest.skip(reason='ASTRA_CUDA not available, skipping 3d test')

    ray_trafo = odl.applications.tomo.RayTransform(space, geometry)
    vol_one = ray_trafo.domain.one()
    data_one = ray_trafo.range.one()

    if ndim == 2:
        # Should raise
        with pytest.raises(NotImplementedError):
            ray_trafo(vol_one)
        with pytest.raises(NotImplementedError):
            ray_trafo.adjoint(data_one)
    elif ndim == 3:
        # Just check that this doesn't crash and computes something nonzero
        data = ray_trafo(vol_one)
        backproj = ray_trafo.adjoint(data_one)
        assert data.norm() > 0
        assert backproj.norm() > 0
    else:
        assert False


def test_shifted_volume(impl, geometry_type, odl_impl_device_pairs):
    """Check that geometry shifts are handled correctly.

    We forward project a square/cube of all ones and check that the
    correct portion of the detector gets nonzero values. In the default
    setup, at angle 0, the source (if existing) is at (0, -s[, 0]), and
    the detector at (0, +d[, 0]) with the positive x axis as (first)
    detector axis. Thus, when shifting enough in the negative x direction,
    the object should be visible at the left half of the detector only.
    A shift in y should not influence the result (much).

    At +90 degrees, a shift in the negative y direction should have the same
    effect.
    """
    apart = odl.nonuniform_partition([0, np.pi / 2, np.pi, 3 * np.pi / 2])
    if geometry_type == 'par2d' and odl.applications.tomo.ASTRA_AVAILABLE:
        ndim = 2
        dpart = odl.uniform_partition(-30, 30, 30)
        geometry = odl.applications.tomo.Parallel2dGeometry(apart, dpart)
    elif geometry_type == 'par3d' and odl.applications.tomo.ASTRA_CUDA_AVAILABLE:
        ndim = 3
        dpart = odl.uniform_partition([-30, -30], [30, 30], (30, 30))
        geometry = odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart)
    if geometry_type == 'cone2d' and odl.applications.tomo.ASTRA_AVAILABLE:
        ndim = 2
        dpart = odl.uniform_partition(-30, 30, 30)
        geometry = odl.applications.tomo.FanBeamGeometry(apart, dpart,
                                            src_radius=200, det_radius=100)
    elif geometry_type == 'cone3d' and odl.applications.tomo.ASTRA_CUDA_AVAILABLE:
        ndim = 3
        dpart = odl.uniform_partition([-30, -30], [30, 30], (30, 30))
        geometry = odl.applications.tomo.ConeBeamGeometry(apart, dpart,
                                             src_radius=200, det_radius=100)
    else:
        pytest.skip('no projector available for geometry type')

    impl, device = odl_impl_device_pairs

    min_pt = np.array([-5.0] * ndim)
    max_pt = np.array([5.0] * ndim)
    shift_len = 6  # enough to move the projection to one side of the detector

    # Shift along axis 0
    shift = np.zeros(ndim)
    shift[0] = -shift_len

    # Generate 4 projections with 90 degrees increment
    space = odl.uniform_discr(min_pt + shift, max_pt + shift, [10] * ndim, dtype='float32', impl=impl, device=device)
    ray_trafo = odl.applications.tomo.RayTransform(space, geometry)
    proj = ray_trafo(space.one())

    # Check that the object is projected to the correct place. With the
    # chosen setup, at least one ray should go through a substantial
    # part of the volume, yielding a value around 10 (=side length).

    # 0 degrees: All on the left
    assert odl.max(proj[0, :15]) > 5
    assert odl.max(proj[0, 15:]) == 0

    # 90 degrees: Left and right
    assert odl.max(proj[1, :15]) > 5
    assert odl.max(proj[1, 15:]) > 5

    # 180 degrees: All on the right
    assert odl.max(proj[2, :15]) == 0
    assert odl.max(proj[2, 15:]) > 5

    # 270 degrees: Left and right
    assert odl.max(proj[3, :15]) > 5
    assert odl.max(proj[3, 15:]) > 5

    # Do the same for axis 1
    shift = np.zeros(ndim)
    shift[1] = -shift_len

    space = odl.uniform_discr(min_pt + shift, max_pt + shift, [10] * ndim, dtype='float32', impl=impl, device=device)
    ray_trafo = odl.applications.tomo.RayTransform(space, geometry)
    proj = ray_trafo(space.one())

    # 0 degrees: Left and right
    assert odl.max(proj[0, :15]) > 5
    assert odl.max(proj[0, 15:]) > 5

    # 90 degrees: All on the left
    assert odl.max(proj[1, :15]) > 5
    assert odl.max(proj[1, 15:]) == 0

    # 180 degrees: Left and right
    assert odl.max(proj[2, :15]) > 5
    assert odl.max(proj[2, 15:]) > 5

    # 270 degrees: All on the right
    assert odl.max(proj[3, :15]) == 0
    assert odl.max(proj[3, 15:]) > 5


def test_detector_shifts_2d(impl, odl_impl_device_pairs):
    """Check that detector shifts are handled correctly.

    We forward project a cubic phantom and check that ray transform
    and back-projection with and without detector shifts are
    numerically close (the error depends on domain discretization).
    """
    astra_impl = impl
    tspace_impl, device = odl_impl_device_pairs
    if not odl.applications.tomo.ASTRA_AVAILABLE:
        pytest.skip(reason='ASTRA not available, skipping 2d test')

    if astra_impl == 'astra_cuda':
        pytest.skip(reason='This test produces a known error for astra_cuda, passing')

    if impl == 'skimage':
        pytest.skip(f'Skimage backend not available with pytofor Fan-Beam Geometry') 
    
    d = 10
    space = odl.uniform_discr([-1] * 2, [1] * 2, [d] * 2, dtype='float32', impl=tspace_impl, device=device)
    ns = space.array_namespace
    phantom = odl.core.phantom.cuboid(space, [-1 / 3] * 2, [1 / 3] * 2)

    full_angle = 2 * np.pi
    n_angles = 2 * 10
    src_rad = 2
    det_rad = 2
    apart = odl.uniform_partition(0, full_angle, n_angles)
    dpart = odl.uniform_partition(-4, 4, 8 * d)
    geom = odl.applications.tomo.FanBeamGeometry(apart, dpart, src_rad, det_rad)
    k = 3
    shift = k * dpart.cell_sides[0]
    geom_shift = odl.applications.tomo.FanBeamGeometry(
        apart, dpart, src_rad, det_rad,
        det_shift_func=lambda angle: [0.0, shift]
    )

    assert all_almost_equal(geom.angles, geom_shift.angles)
    angles = geom.angles
    assert all_almost_equal(geom.src_position(angles),
                            geom_shift.src_position(angles))
    assert all_almost_equal(geom.det_axis(angles),
                            geom_shift.det_axis(angles))
    assert all_almost_equal(geom.det_refpoint(angles),
                            geom_shift.det_refpoint(angles)
                            + shift * geom_shift.det_axis(angles))

    # check ray transform
    op = odl.applications.tomo.RayTransform(space, geom, impl=impl)
    op_shift = odl.applications.tomo.RayTransform(space, geom_shift, impl=astra_impl)
    y = op(phantom).asarray()
    y_shift = op_shift(phantom).asarray()
    # projection on the shifted detector is shifted regular projection
    data_error = ns.max(ns.abs(y[:, :-k] - y_shift[:, k:]))
    assert data_error < space.cell_volume

    # check back-projection
    im = op.adjoint(y).asarray()
    im_shift = op_shift.adjoint(y_shift).asarray()
    error = ns.abs(im_shift - im)
    rel_error = ns.max(error[im > 0] / im[im > 0])
    assert rel_error < space.cell_volume


def test_source_shifts_2d(odl_impl_device_pairs):
    """Check that source shifts are handled correctly.

    We forward project a Shepp-Logan phantom and check that reconstruction
    with flying focal spot is equal to a sum of reconstructions with two
    geometries which mimic ffs by using initial angular offsets and
    detector shifts
    """
    tspace_impl, device = odl_impl_device_pairs

    if not odl.applications.tomo.ASTRA_AVAILABLE:
        pytest.skip(reason='ASTRA required but not available')

    if tspace_impl == 'pytorch' and impl == 'skimage':
        pytest.skip(f'Skimage backend not available with pytorch') 

    d = 10
    space = odl.uniform_discr([-1] * 2, [1] * 2, [d] * 2, dtype='float32', impl=tspace_impl, device=device)
    ns = space.array_namespace
    phantom = odl.core.phantom.cuboid(space, [-1 / 3] * 2, [1 / 3] * 2)

    full_angle = 2 * np.pi
    n_angles = 2 * 10
    src_rad = 2
    det_rad = 2
    apart = odl.uniform_partition(0, full_angle, n_angles)
    dpart = odl.uniform_partition(-4, 4, 8 * d)
    # Source positions with flying focal spot should correspond to
    # source positions of 2 geometries with different starting positions
    shift1 = np.array([0.0, -0.3])
    shift2 = np.array([0.0, 0.3])
    init = np.array([1, 0], dtype=np.float32)
    det_init = np.array([0, -1], dtype=np.float32)

    ffs = partial(odl.applications.tomo.flying_focal_spot,
                  apart=apart,
                  shifts=[shift1, shift2])
    geom_ffs = odl.applications.tomo.FanBeamGeometry(apart, dpart,
                                        src_rad, det_rad,
                                        src_to_det_init=init,
                                        det_axis_init=det_init,
                                        src_shift_func=ffs,
                                        det_shift_func=ffs)
    # angles must be shifted to match discretization of apart
    ang1 = -full_angle / (n_angles * 2)
    apart1 = odl.uniform_partition(ang1, full_angle + ang1, n_angles // 2)
    ang2 = full_angle / (n_angles * 2)
    apart2 = odl.uniform_partition(ang2, full_angle + ang2, n_angles // 2)

    init1 = init + np.array([0, shift1[1]]) / (src_rad + shift1[0])
    init2 = init + np.array([0, shift2[1]]) / (src_rad + shift2[0])
    # radius also changes when a shift is applied
    src_rad1 = np.linalg.norm(np.array([src_rad, 0]) + shift1)
    src_rad2 = np.linalg.norm(np.array([src_rad, 0]) + shift2)
    det_rad1 = np.linalg.norm(
        np.array([det_rad, shift1[1] / src_rad * det_rad]))
    det_rad2 = np.linalg.norm(
        np.array([det_rad, shift2[1] / src_rad * det_rad]))
    geom1 = odl.applications.tomo.FanBeamGeometry(apart1, dpart,
                                     src_rad1, det_rad1,
                                     src_to_det_init=init1,
                                     det_axis_init=det_init)
    geom2 = odl.applications.tomo.FanBeamGeometry(apart2, dpart,
                                     src_rad2, det_rad2,
                                     src_to_det_init=init2,
                                     det_axis_init=det_init)

    # check ray transform
    op_ffs = odl.applications.tomo.RayTransform(space, geom_ffs)
    op1 = odl.applications.tomo.RayTransform(space, geom1)
    op2 = odl.applications.tomo.RayTransform(space, geom2)
    y_ffs = op_ffs(phantom)
    y1 = op1(phantom).asarray()
    y2 = op2(phantom).asarray()
    assert all_almost_equal(y_ffs[::2], y1)
    assert all_almost_equal(y_ffs[1::2], y2)

    # check back-projection
    im = op_ffs.adjoint(y_ffs).asarray()
    im1 = op1.adjoint(y1).asarray()
    im2 = op2.adjoint(y2).asarray()
    im_combined = (im1 + im2) / 2
    rel_error = ns.abs((im - im_combined)[im > 0] / im[im > 0])
    assert ns.max(rel_error) < 1e-6


def test_detector_shifts_3d(impl, odl_impl_device_pairs):
    """Check that detector shifts are handled correctly.

    We forward project a cubic phantom and check that ray transform
    and back-projection with and without detector shifts are
    numerically close (the error depends on domain discretization).
    """
    tspace_impl, device = odl_impl_device_pairs
    if not odl.applications.tomo.ASTRA_CUDA_AVAILABLE:
        pytest.skip(reason='ASTRA CUDA required but not available')

    d = 100
    space = odl.uniform_discr([-1] * 3, [1] * 3, [d] * 3, dtype='float32', impl=tspace_impl, device=device)
    ns = space.array_namespace
    phantom = odl.core.phantom.cuboid(space, [-1 / 3] * 3, [1 / 3] * 3)

    full_angle = 2 * np.pi
    n_angles = 2 * 100
    src_rad = 2
    det_rad = 2
    apart = odl.uniform_partition(0, full_angle, n_angles)
    dpart = odl.uniform_partition([-4] * 2, [4] * 2, [8 * d] * 2)
    geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart, src_rad, det_rad)
    k = 3
    l = 2
    shift = np.array([0, k, l]) * dpart.cell_sides[0]
    geom_shift = odl.applications.tomo.ConeBeamGeometry(apart, dpart, src_rad, det_rad,
                                           det_shift_func=lambda angle: shift)

    angles = geom.angles

    assert all_almost_equal(angles, geom_shift.angles)
    assert all_almost_equal(geom.src_position(angles),
                            geom_shift.src_position(angles))
    assert all_almost_equal(geom.det_axes(angles),
                            geom_shift.det_axes(angles))
    assert all_almost_equal(geom.det_refpoint(angles),
                            geom_shift.det_refpoint(angles)
                            + geom_shift.det_axes(angles)[:, 0] * shift[1]
                            - geom_shift.det_axes(angles)[:, 1] * shift[2])

    # check forward pass
    op = odl.applications.tomo.RayTransform(space, geom)
    op_shift = odl.applications.tomo.RayTransform(space, geom_shift)
    y = op(phantom).asarray()
    y_shift = op_shift(phantom).asarray()
    data_error = ns.max(ns.abs(y[:, :-k, l:] - y_shift[:, k:, :-l]))
    assert data_error < 1e-3

    # check back-projection
    im = op.adjoint(y).asarray()
    im_shift = op_shift.adjoint(y_shift).asarray()
    error = ns.max(ns.abs(im_shift - im))
    assert error < 1e-3


def test_source_shifts_3d(odl_impl_device_pairs):
    """Check that source shifts are handled correctly.

    We forward project a Shepp-Logan phantom and check that reconstruction
    with flying focal spot is equal to a sum of reconstructions with two
    geometries which mimic ffs by using initial angular offsets and
    detector shifts
    """
    impl, device = odl_impl_device_pairs
    if not odl.applications.tomo.ASTRA_CUDA_AVAILABLE:
        pytest.skip(reason='ASTRA_CUDA not available, skipping 3d test')

    d = 10
    space = odl.uniform_discr([-1] * 3, [1] * 3, [d] * 3, dtype='float32', impl=impl, device=device)
    ns = space.array_namespace
    phantom = odl.core.phantom.cuboid(space, [-1 / 3] * 3, [1 / 3] * 3)

    full_angle = 2 * np.pi
    n_angles = 2 * 10
    apart = odl.uniform_partition(0, full_angle, n_angles)
    dpart = odl.uniform_partition([-4] * 2, [4] * 2, [8 * d] * 2)
    src_rad = 2
    det_rad = 2
    pitch = 0.2
    # Source positions with flying focal spot should correspond to
    # source positions of 2 geometries with different starting positions
    shift1 = np.array([0.0, -0.2, 0.1])
    shift2 = np.array([0.0, 0.2, -0.1])
    init = np.array([1, 0, 0], dtype=np.float32)
    det_init = np.array([[0, -1, 0], [0, 0, 1]], dtype=np.float32)
    ffs = partial(odl.applications.tomo.flying_focal_spot,
                  apart=apart,
                  shifts=[shift1, shift2])
    geom_ffs = odl.applications.tomo.ConeBeamGeometry(apart, dpart,
                                         src_rad, det_rad,
                                         src_to_det_init=init,
                                         det_axes_init=det_init,
                                         src_shift_func=ffs,
                                         det_shift_func=ffs,
                                         pitch=pitch)
    # angles must be shifted to match discretization of apart
    ang1 = -full_angle / (n_angles * 2)
    apart1 = odl.uniform_partition(ang1, full_angle + ang1, n_angles // 2)
    ang2 = full_angle / (n_angles * 2)
    apart2 = odl.uniform_partition(ang2, full_angle + ang2, n_angles // 2)

    init1 = init + np.array([0, shift1[1], 0]) / (src_rad + shift1[0])
    init2 = init + np.array([0, shift2[1], 0]) / (src_rad + shift2[0])
    # radius also changes when a shift is applied
    src_rad1 = np.linalg.norm(np.array([src_rad + shift1[0], shift1[1], 0]))
    src_rad2 = np.linalg.norm(np.array([src_rad + shift2[0], shift2[1], 0]))
    det_rad1 = np.linalg.norm(
        np.array([det_rad, det_rad / src_rad * shift1[1], 0]))
    det_rad2 = np.linalg.norm(
        np.array([det_rad, det_rad / src_rad * shift2[1], 0]))
    geom1 = odl.applications.tomo.ConeBeamGeometry(apart1, dpart, src_rad1, det_rad1,
                                      src_to_det_init=init1,
                                      det_axes_init=det_init,
                                      offset_along_axis=shift1[2],
                                      pitch=pitch)
    geom2 = odl.applications.tomo.ConeBeamGeometry(apart2, dpart, src_rad2, det_rad2,
                                      src_to_det_init=init2,
                                      det_axes_init=det_init,
                                      offset_along_axis=shift2[2],
                                      pitch=pitch)

    assert all_almost_equal(geom_ffs.src_position(geom_ffs.angles)[::2],
                            geom1.src_position(geom1.angles))
    assert all_almost_equal(geom_ffs.src_position(geom_ffs.angles)[1::2],
                            geom2.src_position(geom2.angles))

    assert all_almost_equal(geom_ffs.det_refpoint(geom_ffs.angles)[::2],
                            geom1.det_refpoint(geom1.angles))
    assert all_almost_equal(geom_ffs.det_refpoint(geom_ffs.angles)[1::2],
                            geom2.det_refpoint(geom2.angles))

    assert all_almost_equal(geom_ffs.det_axes(geom_ffs.angles)[::2],
                            geom1.det_axes(geom1.angles))
    assert all_almost_equal(geom_ffs.det_axes(geom_ffs.angles)[1::2],
                            geom2.det_axes(geom2.angles))

    op_ffs = odl.applications.tomo.RayTransform(space, geom_ffs)
    op1 = odl.applications.tomo.RayTransform(space, geom1)
    op2 = odl.applications.tomo.RayTransform(space, geom2)
    y_ffs = op_ffs(phantom)
    y1 = op1(phantom)
    y2 = op2(phantom)
    assert all_almost_equal(odl.mean(y_ffs[::2], axis=(1, 2)),
                            odl.mean(y1, axis=(1, 2)))
    assert all_almost_equal(odl.mean(y_ffs[1::2], axis=(1, 2)),
                            odl.mean(y2, axis=(1, 2)))
    im = op_ffs.adjoint(y_ffs).asarray()
    im_combined = (op1.adjoint(y1).asarray() + op2.adjoint(y2).asarray())
    # the scaling is a bit off for older versions of astra
    im_combined = im_combined / ns.sum(im_combined) * ns.sum(im)
    rel_error = ns.abs((im - im_combined)[im > 0] / im[im > 0])
    assert ns.max(rel_error) < 1e-6


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

__all__ = ('param_opt', )

from .param_opt import *
__all__ += param_opt.__all__



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division

import inspect
import sys

import numpy as np
import pytest

import odl
from odl import (
    FunctionalLeftVectorMult, MatrixOperator, OpDomainError, Operator,
    OperatorComp, OperatorLeftScalarMult, OperatorLeftVectorMult,
    OperatorRightScalarMult, OperatorRightVectorMult, OperatorSum,
    OpRangeError, OpTypeError)
from odl.core.operator.operator import _dispatch_call_args, _function_signature
from odl.core.util.testutils import (
    all_almost_equal, noise_element, noise_elements, simple_fixture)

from odl.core.array_API_support.utils import get_array_and_backend, lookup_array_backend

try:
    getargspec = inspect.getfullargspec
except AttributeError:
    getargspec = inspect.getargspec


# --- Fixtures --- #


# This fixture is intended to activate testing of operator evaluation
# with aliased input and output, which is only possible if domain == range.
dom_eq_ran = simple_fixture('dom_eq_ran', [True, False])


# --- Auxiliary --- #


class MultiplyAndSquareOp(Operator):
    """Example of a nonlinear operator, x --> (mat*x)**2."""

    def __init__(self, matrix, domain=None, range=None):
        matrix, backend = get_array_and_backend(matrix)

        dom = (odl.rn(matrix.shape[1], impl=backend.impl, device=str(matrix.device), dtype=matrix.dtype)
               if domain is None else domain)
        
        ran = (odl.rn(matrix.shape[0], impl=backend.impl, device=str(matrix.device), dtype=matrix.dtype)
               if range is None else range)

        super(MultiplyAndSquareOp, self).__init__(dom, ran)
        self.matrix = matrix

    def _call(self, x, out=None):
        out_of_place = False
        if out is None:
            out_of_place = True
            out = self.range.element()
        out[:] = self.matrix @ x.data
        out **= 2
        if out_of_place:
            return out

    def derivative(self, x):
        return 2 * odl.MatrixOperator(self.matrix)

    def __str__(self):
        return "MaS: " + str(self.matrix) + " ** 2"

def mult_sq_np(mat, x):
    """NumPy reference implementation of MultiplyAndSquareOp."""
    mat, backend_mat = get_array_and_backend(mat)
    x, backend_x = get_array_and_backend(x)
    assert backend_mat == backend_x 
    return (x @ mat.T) ** 2


def check_call(operator, x, expected):
    """Assert that operator(point) == expected."""
    # Out-of-place check
    assert all_almost_equal(operator(x), expected)

    # In-place check, no aliasing
    out = operator.range.element()
    operator(x, out=out)
    assert all_almost_equal(out, expected)

    # In-place check, aliased
    if operator.domain == operator.range:
        y = x.copy()
        operator(y, out=y)
        assert all_almost_equal(y, expected)


# --- Unit tests --- #

@pytest.fixture(scope="module", ids=['True', 'False'], params=[True, False])
def dom_eq_ran_mat(request, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    dom_eq_ran = request
    if dom_eq_ran:
        shape = (3,3)
    else:
        shape = (3,4)
    space = odl.rn(shape, impl=impl, device=device)
    mat, _ = noise_elements(space)
    return mat

def test_operator_call__(dom_eq_ran_mat):
    """Check operator evaluation against NumPy reference."""
    op = MultiplyAndSquareOp(dom_eq_ran_mat)
    xarr, x = noise_elements(op.domain)
    assert all_almost_equal(op(x), mult_sq_np(dom_eq_ran_mat, xarr))

def test_operator_call_in_place_wrong_return(odl_impl_device_pairs):
    """Test that operator with out parameter actually returns out."""
    impl, device = odl_impl_device_pairs
    class BadInplaceOperator(odl.Operator):
        def _call(self, x, out):
            # badly implemented operator
            out = 42
            return out

    space = odl.rn(3, impl=impl, device=device)
    op = BadInplaceOperator(space, space)

    with pytest.raises(ValueError):
        op(space.zero())

    with pytest.raises(ValueError):
        out = space.zero()
        op(space.zero(), out=out)


def test_operator_sum(dom_eq_ran_mat):
    """Check operator sum against NumPy reference."""
    mat1 = dom_eq_ran_mat
    mat2 = dom_eq_ran_mat + 0.5
    op1 = MultiplyAndSquareOp(mat1)
    op2 = MultiplyAndSquareOp(mat2)
    xarr, x = noise_elements(op1.domain)

    # Explicit instantiation
    sum_op = OperatorSum(op1, op2)
    assert not sum_op.is_linear
    check_call(sum_op, x, mult_sq_np(mat1, xarr) + mult_sq_np(mat2, xarr))

    # Using operator overloading
    check_call(op1 + op2, x, mult_sq_np(mat1, xarr) + mult_sq_np(mat2, xarr))

    # Verify that unmatched operator domains fail
    op_wrong_dom = MultiplyAndSquareOp(mat1[:, :-1])
    with pytest.raises(OpTypeError):
        OperatorSum(op1, op_wrong_dom)

    # Verify that unmatched operator ranges fail
    op_wrong_ran = MultiplyAndSquareOp(mat1[:-1, :])
    with pytest.raises(OpTypeError):
        OperatorSum(op1, op_wrong_ran)


def test_operator_scaling(dom_eq_ran_mat):
    """Check operator scaling against NumPy reference."""
    mat = dom_eq_ran_mat

    op = MultiplyAndSquareOp(mat)
    xarr, x = noise_elements(op.domain)

    # Test a range of scalars (scalar multiplication could implement
    # optimizations for (-1, 0, 1)).
    scalars = [-1.432, -1, 0, 1, 3.14]
    for scalar in scalars:
        lscaled = OperatorLeftScalarMult(op, scalar)
        rscaled = OperatorRightScalarMult(op, scalar)

        assert not lscaled.is_linear
        assert not rscaled.is_linear

        # Explicit
        check_call(lscaled, x, scalar * mult_sq_np(mat, xarr))
        check_call(rscaled, x, mult_sq_np(mat, scalar * xarr))

        # Using operator overloading
        check_call(scalar * op, x, scalar * mult_sq_np(mat, xarr))
        check_call(op * scalar, x, mult_sq_np(mat, scalar * xarr))

    # Fail when scaling by wrong scalar type (complex number)
    wrongscalars = [1j, [1, 2], (1, 2)]
    for wrongscalar in wrongscalars:
        with pytest.raises(TypeError):
            OperatorLeftScalarMult(op, wrongscalar)

        with pytest.raises(TypeError):
            OperatorRightScalarMult(op, wrongscalar)

        with pytest.raises(TypeError):
            op * wrongscalar

        with pytest.raises(TypeError):
            wrongscalar * op


def test_operator_vector_mult(dom_eq_ran_mat):
    """Check operator-vector multiplication against NumPy reference."""
    mat = dom_eq_ran_mat

    op = MultiplyAndSquareOp(mat)
    right = op.domain.element(np.arange(op.domain.size))
    left = op.range.element(np.arange(op.range.size))
    xarr, x = noise_elements(op.domain)

    right_as_array = right.asarray()
    left_as_array  = left.asarray()

    rmult_op = OperatorRightVectorMult(op, right)
    lmult_op = OperatorLeftVectorMult(op, left)
    assert not rmult_op.is_linear
    assert not lmult_op.is_linear

    check_call(rmult_op, x, mult_sq_np(mat, right_as_array * xarr))
    check_call(lmult_op, x, left_as_array * mult_sq_np(mat, xarr))

    # Using operator overloading
    check_call(op @ right, x, mult_sq_np(mat, right_as_array * xarr))
    check_call(left @ op, x, left_as_array * mult_sq_np(mat, xarr))


def test_operator_composition(dom_eq_ran_mat):
    """Check operator composition against NumPy reference."""
    mat1 = dom_eq_ran_mat
    mat2 = dom_eq_ran_mat + 0.5

    op1 = MultiplyAndSquareOp(mat1)
    op2 = MultiplyAndSquareOp(mat2)
    xarr, x = noise_elements(op2.domain)

    comp_op = OperatorComp(op1, op2)
    assert not comp_op.is_linear
    check_call(comp_op, x, mult_sq_np(mat1, mult_sq_np(mat2, xarr)))

    # Verify that incorrect order fails
    if not dom_eq_ran:
        with pytest.raises(OpTypeError):
            OperatorComp(op2, op1)


def test_linear_operator_call(dom_eq_ran_mat):
    """Check call of a linear operator against NumPy, and ``is_linear``."""
    mat = dom_eq_ran_mat

    op = MatrixOperator(mat)
    _, backend = get_array_and_backend(mat)
    assert op.is_linear
    xarr, x = noise_elements(op.domain)

    check_call(op, x, backend.array_namespace.matmul(mat, xarr))


def test_linear_operator_adjoint(dom_eq_ran_mat):
    """Check adjoint of a linear operator against NumPy."""
    mat = dom_eq_ran_mat

    op = MatrixOperator(mat)
    _, backend = get_array_and_backend(mat)
    xarr, x = noise_elements(op.range)
    check_call(op.adjoint, x, backend.array_namespace.matmul(mat.T, xarr))


def test_linear_operator_addition(dom_eq_ran_mat):
    """Check call and adjoint of a sum of linear operators."""
    mat1 = dom_eq_ran_mat
    mat2 = dom_eq_ran_mat + 0.5
    _, backend = get_array_and_backend(mat1)
    ns = backend.array_namespace
    op1 = MatrixOperator(mat1)
    op2 = MatrixOperator(mat2)
    xarr, x = noise_elements(op1.domain)
    yarr, y = noise_elements(op1.range)

    # Explicit instantiation
    sum_op = OperatorSum(op1, op2)
    assert sum_op.is_linear
    assert sum_op.adjoint.is_linear
    check_call(sum_op, x, ns.matmul(mat1, xarr) + ns.matmul(mat2, xarr))
    check_call(sum_op.adjoint, y, ns.matmul(mat1.T, yarr) + ns.matmul(mat2.T, yarr))

    # Using operator overloading
    check_call(op1 + op2, x, ns.matmul(mat1, xarr) + ns.matmul(mat2, xarr))
    check_call((op1 + op2).adjoint,
               y, ns.matmul(mat1.T, yarr) + ns.matmul(mat2.T, yarr))


def test_linear_operator_scaling(dom_eq_ran_mat):
    """Check call and adjoint of a scaled linear operator."""
    mat = dom_eq_ran_mat

    op = MatrixOperator(mat)
    _, backend = get_array_and_backend(mat)
    ns = backend.array_namespace

    xarr, x = noise_elements(op.domain)
    yarr, y = noise_elements(op.range)
    
    # Test a range of scalars (scalar multiplication could implement
    # optimizations for (-1, 0, 1).
    scalars = [-1.432, -1, 0, 1, 3.14]
    for scalar in scalars:
        # Explicit instantiation
        scaled_op = OperatorRightScalarMult(op, scalar)
        assert scaled_op.is_linear
        assert scaled_op.adjoint.is_linear
        check_call(scaled_op, x, scalar * ns.matmul(mat, xarr))
        check_call(scaled_op.adjoint, y, scalar * ns.matmul(mat.T, yarr))

        # Using operator overloading
        check_call(scalar * op, x, scalar * ns.matmul(mat, xarr))
        check_call(op * scalar, x, scalar * ns.matmul(mat, xarr))
        check_call((scalar * op).adjoint, y, scalar * ns.matmul(mat.T, yarr))
        check_call((op * scalar).adjoint, y, scalar * ns.matmul(mat.T, yarr))


def test_linear_right_vector_mult(dom_eq_ran_mat):
    """Check call and adjoint of linear operator x vector."""
    mat = dom_eq_ran_mat

    op = MatrixOperator(mat)
    _, backend = get_array_and_backend(mat)
    ns = backend.array_namespace
    (xarr, mul_arr), (x, mul) = noise_elements(op.domain, n=2)
    yarr, y = noise_elements(op.range)

    # Explicit instantiation
    rmult_op = OperatorRightVectorMult(op, mul)
    assert rmult_op.is_linear
    assert rmult_op.adjoint.is_linear
    check_call(rmult_op, x, ns.matmul(mat, mul_arr * xarr))
    check_call(rmult_op.adjoint, y, mul_arr * ns.matmul(mat.T, yarr))

    # Using operator overloading
    check_call(op * mul, x, ns.matmul(mat, mul_arr * xarr))
    check_call((op * mul).adjoint, y, mul_arr * ns.matmul(mat.T, yarr))


def test_linear_left_vector_mult(dom_eq_ran_mat):
    """Check call and adjoint of vector x linear operator."""
    mat = dom_eq_ran_mat
    
    op = MatrixOperator(mat)
    _, backend = get_array_and_backend(mat)
    ns = backend.array_namespace
    xarr, x = noise_elements(op.domain)
    (yarr, mul_arr), (y, mul) = noise_elements(op.range, n=2)

    # Explicit instantiation
    lmult_op = OperatorLeftVectorMult(op, mul)
    assert lmult_op.is_linear
    assert lmult_op.adjoint.is_linear
    check_call(lmult_op, x, mul_arr * ns.matmul(mat, xarr))
    check_call(lmult_op.adjoint, y, ns.matmul(mat.T, mul_arr * yarr))

    # Using operator overloading
    check_call(mul @ op, x, mul_arr * ns.matmul(mat, xarr))
    check_call((mul @ op).adjoint, y, ns.matmul(mat.T, mul_arr * yarr))


def test_linear_operator_composition(dom_eq_ran_mat):
    """Check call and adjoint of linear operator composition."""
    mat1 = dom_eq_ran_mat
    mat2 = dom_eq_ran_mat + 0.5

    op1 = MatrixOperator(mat1)
    op2 = MatrixOperator(mat2)
    _, backend = get_array_and_backend(mat1)
    ns = backend.array_namespace
    xarr, x = noise_elements(op2.domain)
    yarr, y = noise_elements(op1.range)

    # Explicit instantiation
    comp_op = OperatorComp(op1, op2)
    assert comp_op.is_linear
    assert comp_op.adjoint.is_linear
    check_call(comp_op, x, ns.matmul(mat1, ns.matmul(mat2, xarr)))
    check_call(comp_op.adjoint, y, ns.matmul(mat2.T, ns.matmul(mat1.T, yarr)))

    # Using operator overloading
    check_call(op1 * op2, x, ns.matmul(mat1, ns.matmul(mat2, xarr)))
    check_call((op1 * op2).adjoint, y, ns.matmul(mat2.T, ns.matmul(mat1.T, yarr)))


def test_type_errors(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device = device)
    r4 = odl.rn(4, impl=impl, device = device)
    space = odl.rn((3,3), impl=impl, device=device)

    op = MatrixOperator(space.element())
    r3_elem1 = r3.zero()
    r3_elem2 = r3.zero()
    r4_elem1 = r4.zero()
    r4_elem2 = r4.zero()

    # Verify that correct usage works
    op(r3_elem1, r3_elem2)
    op.adjoint(r3_elem1, r3_elem2)

    # Test that erroneous usage raises
    with pytest.raises(OpDomainError):
        op(r4_elem1)

    with pytest.raises(OpDomainError):
        op.adjoint(r4_elem1)

    with pytest.raises(OpRangeError):
        op(r3_elem1, r4_elem1)

    with pytest.raises(OpRangeError):
        op.adjoint(r3_elem1, r4_elem1)

    with pytest.raises(OpDomainError):
        op(r4_elem1, r3_elem1)

    with pytest.raises(OpDomainError):
        op.adjoint(r4_elem1, r3_elem1)

    with pytest.raises(OpDomainError):
        op(r4_elem1, r4_elem2)

    with pytest.raises(OpDomainError):
        op.adjoint(r4_elem1, r4_elem2)


def test_arithmetic(dom_eq_ran_mat):
    """Test that all standard arithmetic works."""
    mat1 = dom_eq_ran_mat
    mat2 = dom_eq_ran_mat + 1 
    mat3 = dom_eq_ran_mat + 2 
    mat4 = dom_eq_ran_mat + 3

    op = MultiplyAndSquareOp(mat1)
    op2 = MultiplyAndSquareOp(mat2)
    op3 = MultiplyAndSquareOp(mat3)
    op4 = MultiplyAndSquareOp(mat4)
    # Create elements needed for later
    x = noise_element(op.domain)
    y = noise_element(op.domain)
    z = noise_element(op.range)
    scalar = np.pi

    # Simple tests here, more in depth comes later
    check_call(+op, x, op(x))
    check_call(-op, x, -op(x))
    check_call(scalar * op, x, scalar * op(x))
    check_call(scalar * (scalar * op), x, scalar**2 * op(x))
    check_call(op * scalar, x, op(scalar * x))
    check_call((op * scalar) * scalar, x, op(scalar**2 * x))
    check_call(op + op2, x, op(x) + op2(x))
    check_call(op - op2, x, op(x) - op2(x))
    check_call(op @ op3, x, op(op3(x)))
    check_call(op4 @ op, x, op4(op(x)))
    check_call(z @ op, x, z * op(x))
    check_call(z @ (z @ op), x, (z * z)* op(x))
    check_call(op * y, x, op(x * y))
    check_call((op * y) * y, x, op((y * y) * x))
    check_call(op + z, x, op(x) + z)
    check_call(op - z, x, op(x) - z)
    check_call(z + op, x, z + op(x))
    check_call(z - op, x, z - op(x))
    check_call(op + scalar, x, op(x) + scalar)
    check_call(op - scalar, x, op(x) - scalar)
    check_call(scalar + op, x, scalar + op(x))
    check_call(scalar - op, x, scalar - op(x))


def test_operator_pointwise_product(dom_eq_ran_mat):
    """Check call and adjoint of operator pointwise multiplication."""
    mat1 = dom_eq_ran_mat
    mat2 = dom_eq_ran_mat + 1 

    op1 = MultiplyAndSquareOp(mat1)
    op2 = MultiplyAndSquareOp(mat2)
    x = noise_element(op1.domain)

    prod_op = odl.OperatorPointwiseProduct(op1, op2)

    # Evaluate
    expected = op1(x) * op2(x)
    check_call(prod_op, x, expected)

    # Derivative
    y = noise_element(op1.domain)
    expected = (op1.derivative(x)(y) * op2(x) +
                op2.derivative(x)(y) * op1(x))
    prod_deriv_op = prod_op.derivative(x)
    assert prod_deriv_op.is_linear
    check_call(prod_deriv_op, y, expected)

    # Derivative Adjoint
    z = noise_element(op1.range)
    expected = (op1.derivative(x).adjoint(z * op2(x)) +
                op2.derivative(x).adjoint(z * op1(x)))
    prod_deriv_adj_op = prod_deriv_op.adjoint
    assert prod_deriv_adj_op.is_linear
    check_call(prod_deriv_adj_op, z, expected)


# FUNCTIONAL TEST
class SumFunctional(Operator):

    """Sum of elements."""

    def __init__(self, domain):
        super(SumFunctional, self).__init__(domain, domain.field, linear=True)

    def _call(self, x):
        return odl.sum(x)

    @property
    def adjoint(self):
        return ConstantVector(self.domain)


class ConstantVector(Operator):

    """Vector times a scalar."""

    def __init__(self, domain):
        super(ConstantVector, self).__init__(domain.field, domain, linear=True)

    def _call(self, x):
        return self.range.element(np.ones(self.range.size) * x)

    @property
    def adjoint(self):
        return SumFunctional(self.range)


def test_functional(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)
    x = r3.element([1, 2, 3])

    op = SumFunctional(r3)

    assert op(x) == 6


def test_functional_out(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)
    x = r3.element([1, 2, 3])

    op = SumFunctional(r3)

    # No out parameter allowed with functionals
    out = 0
    with pytest.raises(TypeError):
        op(x, out=out)


def test_functional_adjoint(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)

    op = SumFunctional(r3)

    assert op.adjoint(3) == r3.element([3, 3, 3])

    x = r3.element([1, 2, 3])
    assert op.adjoint.adjoint(x) == op(x)


def test_functional_addition(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)

    op = SumFunctional(r3)
    op2 = SumFunctional(r3)
    x = r3.element([1, 2, 3])
    y = 1

    # Explicit instantiation
    C = OperatorSum(op, op2)

    assert C.is_linear
    assert C.adjoint.is_linear

    assert C(x) == 2 * odl.sum(x)

    # Test adjoint
    assert all_almost_equal(C.adjoint(y), y * 2 * r3.one())
    assert all_almost_equal(C.adjoint.adjoint(x), C(x))

    # Using operator overloading
    assert (op + op2)(x) == 2 * odl.sum(x)
    assert all_almost_equal((op + op2).adjoint(y), y * 2 * r3.one())


def test_functional_scale(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)

    op = SumFunctional(r3)
    x = r3.element([1, 2, 3])
    y = 1

    # Test a range of scalars (scalar multiplication could implement
    # optimizations for (-1, 0, 1).
    scalars = [-1.432, -1, 0, 1, 3.14]
    for scalar in scalars:
        C = OperatorRightScalarMult(op, scalar)

        assert C.is_linear
        assert C.adjoint.is_linear

        assert C(x) == scalar * odl.sum(x)
        assert all_almost_equal(C.adjoint(y), scalar * y * r3.one())
        assert all_almost_equal(C.adjoint.adjoint(x), C(x))

        # Using operator overloading
        assert (scalar * op)(x) == scalar * odl.sum(x)
        assert (op * scalar)(x) == scalar * odl.sum(x)
        assert all_almost_equal((scalar * op).adjoint(y),
                                scalar * y * r3.one())
        assert all_almost_equal((op * scalar).adjoint(y),
                                scalar * y * r3.one())


def test_functional_left_vector_mult(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)
    r4 = odl.rn(4, impl=impl, device=device)

    op = SumFunctional(r3)
    x = r3.element([1, 2, 3])
    y = r4.element([3, 2, 1, 5])

    # Test a range of scalars (scalar multiplication could implement
    # optimizations for (-1, 0, 1).
    C = FunctionalLeftVectorMult(op, y)

    assert C.is_linear
    assert C.adjoint.is_linear

    assert all_almost_equal(C(x), y * odl.sum(x))
    assert all_almost_equal(C.adjoint(y), y.inner(y) * r3.one())
    assert all_almost_equal(C.adjoint.adjoint(x), C(x))

    # Using operator overloading
    assert all_almost_equal((y @ op)(x),
                            y * odl.sum(x))
    assert all_almost_equal((y @ op).adjoint(y),
                            y.inner(y) * r3.one())


def test_functional_right_vector_mult(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)

    op = SumFunctional(r3)
    vec = r3.element([1, 2, 3])
    x = r3.element([4, 5, 6])
    y = 2.0

    # Test a range of scalars (scalar multiplication could implement
    # optimizations for (-1, 0, 1).
    C = OperatorRightVectorMult(op, vec)

    assert C.is_linear
    assert C.adjoint.is_linear

    assert all_almost_equal(C(x), odl.sum(vec * x))
    assert all_almost_equal(C.adjoint(y), vec * y)
    assert all_almost_equal(C.adjoint.adjoint(x), C(x))

    # Using operator overloading
    assert all_almost_equal((op * vec)(x),
                            odl.sum(vec * x))
    assert all_almost_equal((op * vec).adjoint(y),
                            vec * y)


def test_functional_composition(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)

    op = SumFunctional(r3)
    op2 = ConstantVector(r3)
    x = r3.element([1, 2, 3])
    y = 1

    C = OperatorComp(op2, op)

    assert C.is_linear
    assert C.adjoint.is_linear

    assert all_almost_equal(C(x), odl.sum(x) * r3.one())
    assert all_almost_equal(C.adjoint(x), odl.sum(x) * r3.one())
    assert all_almost_equal(C.adjoint.adjoint(x), C(x))

    # Using operator overloading
    assert (op * op2)(y) == y * 3
    assert (op * op2).adjoint(y) == y * 3
    assert all_almost_equal((op2 * op)(x),
                            odl.sum(x) * r3.one())
    assert all_almost_equal((op2 * op).adjoint(x),
                            odl.sum(x) * r3.one())


class SumSquaredFunctional(Operator):

    """Sum of the squared elements."""

    def __init__(self, domain):
        super(SumSquaredFunctional, self).__init__(
            domain, domain.field, linear=False)

    def _call(self, x):
        return odl.sum(x ** 2)


def test_nonlinear_functional(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)
    x = r3.element([1, 2, 3])

    op = SumSquaredFunctional(r3)

    assert op(x) == pytest.approx(odl.sum(x ** 2))


def test_nonlinear_functional_out(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)
    x = r3.element([1, 2, 3])

    op = SumSquaredFunctional(r3)
    out = op.range.element()

    with pytest.raises(TypeError):
        op(x, out=out)


def test_nonlinear_functional_operators(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3 = odl.rn(3, impl=impl, device=device)
    x = r3.element([1, 2, 3])

    mat = SumSquaredFunctional(r3)
    mat2 = SumFunctional(r3)

    # Sum
    C = mat + mat2

    assert not C.is_linear
    assert C(x) == pytest.approx(mat(x) + mat2(x))

    # Minus
    C = mat - mat2

    assert not C.is_linear
    assert C(x) == pytest.approx(mat(x) - mat2(x))

    # left mul
    C = 2.0 * mat

    assert not C.is_linear
    assert C(x) == pytest.approx(2.0 * mat(x))

    # right mul
    C = mat * 2.0

    assert not C.is_linear
    assert C(x) == pytest.approx(mat(x * 2.0))

    # right divide
    C = mat / 2.0

    assert not C.is_linear
    assert C(x) == pytest.approx(mat(x / 2.0))


# test functions to dispatch
def f1(x):
    """f1(x)
    False, False
    good
    """


def f2(x, y):
    """f2(x, y)
    False, False
    bad
    """


def f3(x, y, out):
    """f3(x, y, out)
    False, True
    bad
    """


def f4(*args):
    """f4(*args)
    False, False
    bad
    """


def f5(out):
    """f5(out)
    True, False
    bad
    """


def f6(**kwargs):
    """f6(**kwargs)
    False, False
    bad
    """


def f7(*args, **kwargs):
    """f7(*args, **kwargs)
    False, False
    bad
    """


def f8(out=None, param=2):
    """f8(out=None, param=2)
    True, True
    bad
    """


def f9(x, out=None):
    """f9(x, out=None)
    True, True
    good
    """


def f10(x, out=None, param=2):
    """f10(x, out=None, param=2)
    True, True
    bad
    """


def f11(x, out, **kwargs):
    """f11(x, out, **kwargs)
    True, False
    good
    """


def f12(x, y, out=None, param=2, *args, **kwargs):
    """f12(x, y, out=None, param=2, *args, **kwargs)
    True, True
    bad
    """


py3 = sys.version_info.major > 2
if py3:
    exec('''
def f13(x, *, out=None, param=2, **kwargs):
    """f13(x, *, out=None, param=2, **kwargs)
    True, True
    good
    """
''')

    exec('''
def f14(x, *y, out=None, param=2, **kwargs):
    """f14(x, *y, out=None, param=2, **kwargs)
    True, True
    bad
    """
''')


func_params = [eval('f{}'.format(i)) for i in range(1, 13)]
if py3:
    func_params += [eval('f{}'.format(i)) for i in range(13, 15)]
func_ids = [' f{} '.format(i) for i in range(1, 13)]
if py3:
    func_ids += [' f{} '.format(i) for i in range(13, 15)]
func_fixture = pytest.fixture(scope="module", ids=func_ids, params=func_params)


@func_fixture
def func(request):
    return request.param


def test_function_signature(func):

    true_sig = func.__doc__.splitlines()[0].strip()
    sig = _function_signature(func)
    assert true_sig == sig


def test_dispatch_call_args(func):
    # Unbound functions
    true_has, true_opt = eval(func.__doc__.splitlines()[1].strip())
    good = func.__doc__.splitlines()[2].strip() == 'good'

    if good:
        truespec = getargspec(func)
        truespec.args.insert(0, 'self')

        has, opt, spec = _dispatch_call_args(unbound_call=func)

        assert has == true_has
        assert opt == true_opt
        assert spec == truespec
    else:
        with pytest.raises(ValueError):
            _dispatch_call_args(unbound_call=func)


def test_dispatch_call_args_class():

    # Two sneaky classes whose _call method would pass the signature check
    class WithStaticMethod(object):
        @staticmethod
        def _call(x, y, out):
            pass

    class WithClassMethod(object):
        @classmethod
        def _call(cls, x, out=None):
            pass

    with pytest.raises(TypeError):
        _dispatch_call_args(cls=WithStaticMethod)

    with pytest.raises(TypeError):
        _dispatch_call_args(cls=WithClassMethod)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Bindings to the ``pyFFTW`` back-end for Fourier transforms.

The `pyFFTW <https://pyfftw.readthedocs.io>`_  package is a Python
wrapper around the well-known `FFTW <http://fftw.org/>`_ library for fast
Fourier transforms.
"""

from __future__ import print_function, division, absolute_import
from multiprocessing import cpu_count
import numpy as np
from packaging.version import parse as parse_version
import warnings

try:
    import pyfftw
    PYFFTW_AVAILABLE = True
except ImportError:
    PYFFTW_AVAILABLE = False
else:
    if parse_version(pyfftw.__version__) < parse_version('0.10.3'):
        warnings.warn('PyFFTW < 0.10.3 is known to cause problems with some '
                      'ODL functionality, see issue #1002.',
                      RuntimeWarning)

from odl.core.util import (
    is_real_dtype, dtype_repr, complex_dtype, normalized_axes_tuple)

__all__ = ('pyfftw_call', 'PYFFTW_AVAILABLE')


def pyfftw_call(array_in, array_out, direction='forward', axes=None,
                halfcomplex=False, **kwargs):
    """Calculate the DFT with pyfftw.

    The discrete Fourier (forward) transform calcuates the sum::

        f_hat[k] = sum_j( f[j] * exp(-2*pi*1j * j*k/N) )

    where the summation is taken over all indices
    ``j = (j[0], ..., j[d-1])`` in the range ``0 <= j < N``
    (component-wise), with ``N`` being the shape of the input array.

    The output indices ``k`` lie in the same range, except
    for half-complex transforms, where the last axis ``i`` in ``axes``
    is shortened to ``0 <= k[i] < floor(N[i]/2) + 1``.

    In the backward transform, sign of the the exponential argument
    is flipped.

    Parameters
    ----------
    array_in : `numpy.ndarray`
        Array to be transformed
    array_out : `numpy.ndarray`
        Output array storing the transformed values, may be aliased
        with ``array_in``.
    direction : {'forward', 'backward'}, optional
        Direction of the transform
    axes : int or sequence of ints, optional
        Dimensions along which to take the transform. ``None`` means
        using all axes and is equivalent to ``np.arange(ndim)``.
    halfcomplex : bool, optional
        If ``True``, calculate only the negative frequency part along the
        last axis. If ``False``, calculate the full complex FFT.
        This option can only be used with real input data.

    Other Parameters
    ----------------
    fftw_plan : ``pyfftw.FFTW``, optional
        Use this plan instead of calculating a new one. If specified,
        the options ``planning_effort``, ``planning_timelimit`` and
        ``threads`` have no effect.
    planning_effort : str, optional
        Flag for the amount of effort put into finding an optimal
        FFTW plan. See the `FFTW doc on planner flags
        <http://www.fftw.org/fftw3_doc/Planner-Flags.html>`_.
        Available options: {'estimate', 'measure', 'patient', 'exhaustive'}
        Default: 'estimate'
    planning_timelimit : float or ``None``, optional
        Limit planning time to roughly this many seconds.
        Default: ``None`` (no limit)
    threads : int, optional
        Number of threads to use.
        Default: Number of CPUs if the number of data points is larger
        than 4096, else 1.
    normalise_idft : bool, optional
        If ``True``, the result of the backward transform is divided by
        ``1 / N``, where ``N`` is the total number of points in
        ``array_in[axes]``. This ensures that the IDFT is the true
        inverse of the forward DFT.
        Default: ``False``
    import_wisdom : filename or file handle, optional
        File to load FFTW wisdom from. If the file does not exist,
        it is ignored.
    export_wisdom : filename or file handle, optional
        File to append the accumulated FFTW wisdom to

    Returns
    -------
    fftw_plan : ``pyfftw.FFTW``
        The plan object created from the input arguments. It can be
        reused for transforms of the same size with the same data types.
        Note that reuse only gives a speedup if the initial plan
        used a planner flag other than ``'estimate'``.
        If ``fftw_plan`` was specified, the returned object is a
        reference to it.

    Notes
    -----
    * The planning and direction flags can also be specified as
      capitalized and prepended by ``'FFTW_'``, i.e. in the original
      FFTW form.
    * For a ``halfcomplex`` forward transform, the arrays must fulfill
      ``array_out.shape[axes[-1]] == array_in.shape[axes[-1]] // 2 + 1``,
      and vice versa for backward transforms.
    * All planning schemes except ``'estimate'`` require an internal copy
      of the input array but are often several times faster after the
      first call (measuring results are cached). Typically,
      'measure' is a good compromise. If you cannot afford the copy,
      use ``'estimate'``.
    * If a plan is provided via the ``fftw_plan`` parameter, no copy
      is needed internally.
    """
    import pickle

    if not array_in.flags.aligned:
        raise ValueError('input array not aligned')

    if not array_out.flags.aligned:
        raise ValueError('output array not aligned')

    if axes is None:
        axes = tuple(range(array_in.ndim))

    axes = normalized_axes_tuple(axes, array_in.ndim)

    direction = _flag_pyfftw_to_odl(direction)
    fftw_plan_in = kwargs.pop('fftw_plan', None)
    planning_effort = _flag_pyfftw_to_odl(
        kwargs.pop('planning_effort', 'estimate')
    )
    planning_timelimit = kwargs.pop('planning_timelimit', None)
    threads = kwargs.pop('threads', None)
    normalise_idft = kwargs.pop('normalise_idft', False)
    wimport = kwargs.pop('import_wisdom', '')
    wexport = kwargs.pop('export_wisdom', '')

    # Cast input to complex if necessary
    array_in_copied = False
    if is_real_dtype(array_in.dtype) and not halfcomplex:
        # Need to cast array_in to complex dtype
        array_in = array_in.astype(complex_dtype(array_in.dtype))
        array_in_copied = True

    # Do consistency checks on the arguments
    _pyfftw_check_args(array_in, array_out, axes, halfcomplex, direction)

    # Import wisdom if possible
    if wimport:
        try:
            with open(wimport, 'rb') as wfile:
                wisdom = pickle.load(wfile)
        except IOError:
            wisdom = []
        except TypeError:  # Got file handle
            wisdom = pickle.load(wimport)

        if wisdom:
            pyfftw.import_wisdom(wisdom)

    # Copy input array if it hasn't been done yet and the planner is likely
    # to destroy it. If we already have a plan, we don't have to worry.
    planner_destroys = _pyfftw_destroys_input(
        [planning_effort], direction, halfcomplex, array_in.ndim)
    must_copy_array_in = fftw_plan_in is None and planner_destroys

    if must_copy_array_in and not array_in_copied:
        plan_arr_in = np.empty_like(array_in)
        flags = [_flag_odl_to_pyfftw(planning_effort), 'FFTW_DESTROY_INPUT']
    else:
        plan_arr_in = array_in
        flags = [_flag_odl_to_pyfftw(planning_effort)]

    if fftw_plan_in is None:
        if threads is None:
            if plan_arr_in.size <= 4096:  # Trade-off wrt threading overhead
                threads = 1
            else:
                threads = cpu_count()

        fftw_plan = pyfftw.FFTW(
            plan_arr_in, array_out, direction=_flag_odl_to_pyfftw(direction),
            flags=flags, planning_timelimit=planning_timelimit,
            threads=threads, axes=axes)
    else:
        fftw_plan = fftw_plan_in

    if not normalise_idft and direction=='forward':
        fftw_plan(array_in, array_out, normalise_idft=True)
    else:
        fftw_plan(array_in, array_out, normalise_idft=normalise_idft)

    if wexport:
        try:
            with open(wexport, 'ab') as wfile:
                pickle.dump(pyfftw.export_wisdom(), wfile)
        except TypeError:  # Got file handle
            pickle.dump(pyfftw.export_wisdom(), wexport)

    return fftw_plan


def _flag_pyfftw_to_odl(flag):
    return flag.lstrip('FFTW_').lower()


def _flag_odl_to_pyfftw(flag):
    return 'FFTW_' + flag.upper()


def _pyfftw_destroys_input(flags, direction, halfcomplex, ndim):
    """Return ``True`` if FFTW destroys an input array, ``False`` otherwise."""
    if any(flag in flags or _flag_pyfftw_to_odl(flag) in flags
           for flag in ('FFTW_MEASURE', 'FFTW_PATIENT', 'FFTW_EXHAUSTIVE',
                        'FFTW_DESTROY_INPUT')):
        return True
    elif (direction in ('backward', 'FFTW_BACKWARD') and halfcomplex and
          ndim != 1):
        return True
    else:
        return False


def _pyfftw_check_args(arr_in, arr_out, axes, halfcomplex, direction):
    """Raise an error if anything is not ok with in and out."""
    if len(set(axes)) != len(axes):
        raise ValueError('duplicate axes are not allowed')

    if direction == 'forward':
        out_shape = list(arr_in.shape)
        if halfcomplex:
            try:
                out_shape[axes[-1]] = arr_in.shape[axes[-1]] // 2 + 1
            except IndexError:
                raise IndexError('axis index {} out of range for array '
                                 'with {} axes'
                                 ''.format(axes[-1], arr_in.ndim))

        if arr_out.shape != tuple(out_shape):
            raise ValueError('expected output shape {}, got {}'
                             ''.format(tuple(out_shape), arr_out.shape))

        if is_real_dtype(arr_in.dtype):
            out_dtype = complex_dtype(arr_in.dtype)
        elif halfcomplex:
            raise ValueError('cannot combine halfcomplex forward transform '
                             'with complex input')
        else:
            out_dtype = arr_in.dtype

        if arr_out.dtype != out_dtype:
            raise ValueError('expected output dtype {}, got {}'
                             ''.format(dtype_repr(out_dtype),
                                       dtype_repr(arr_out.dtype)))

    elif direction == 'backward':
        in_shape = list(arr_out.shape)
        if halfcomplex:
            try:
                in_shape[axes[-1]] = arr_out.shape[axes[-1]] // 2 + 1
            except IndexError as err:
                raise IndexError('axis index {} out of range for array '
                                 'with {} axes'
                                 ''.format(axes[-1], arr_out.ndim))

        if arr_in.shape != tuple(in_shape):
            raise ValueError('expected input shape {}, got {}'
                             ''.format(tuple(in_shape), arr_in.shape))

        if is_real_dtype(arr_out.dtype):
            in_dtype = complex_dtype(arr_out.dtype)
        elif halfcomplex:
            raise ValueError('cannot combine halfcomplex backward transform '
                             'with complex output')
        else:
            in_dtype = arr_out.dtype

        if arr_in.dtype != in_dtype:
            raise ValueError('expected input dtype {}, got {}'
                             ''.format(dtype_repr(in_dtype),
                                       dtype_repr(arr_in.dtype)))

    else:  # Shouldn't happen
        raise RuntimeError


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests(skip_if=not PYFFTW_AVAILABLE)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import numpy as np
import pytest

import odl
from odl.core.operator.oputils import matrix_representation, power_method_opnorm
from odl.core.operator.pspace_ops import ProductSpaceOperator
from odl.core.util.testutils import all_almost_equal, noise_elements

from odl.core.array_API_support.utils import get_array_and_backend

@pytest.fixture(scope="module", ids=['True', 'False'], params=[True, False])
def dom_eq_ran_mat(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    shape = (3,3)
    space = odl.rn(shape, impl=impl, device=device)
    mat, _ = noise_elements(space)
    return mat


def test_matrix_representation(dom_eq_ran_mat):
    """Verify that the matrix repr returns the correct matrix"""
    Aop = odl.MatrixOperator(dom_eq_ran_mat)
    matrix_repr = matrix_representation(Aop)

    assert all_almost_equal(dom_eq_ran_mat, matrix_repr)


def test_matrix_representation_product_to_lin_space(dom_eq_ran_mat):
    """Verify that the matrix repr works for product spaces.

    Here, since the domain shape ``(2, 3)`` and the range has shape ``(1, 3)``,
    the shape of the matrix representation will be ``(2, 3, 1, 3)``.
    """
    A = dom_eq_ran_mat
    Aop = odl.MatrixOperator(A)

    B = dom_eq_ran_mat+0.1
    Bop = odl.MatrixOperator(B)

    ABop = ProductSpaceOperator([[Aop, Bop]])
    matrix_repr = matrix_representation(ABop)

    assert matrix_repr.shape == (1, 3, 2, 3)

    _, backend = get_array_and_backend(A)
    
    assert backend.to_cpu(
        backend.array_namespace.linalg.norm(A - matrix_repr[0, :, 0, :])) == pytest.approx(0)
    assert backend.to_cpu(
        backend.array_namespace.linalg.norm(B - matrix_repr[0, :, 1, :])) == pytest.approx(0)


def test_matrix_representation_lin_space_to_product(dom_eq_ran_mat):
    """Verify that the matrix repr works for product spaces.

    Here, since the domain shape ``(1, 3)`` and the range has shape ``(2, 3)``,
    the shape of the matrix representation will be ``(2, 3, 1, 3)``.
    """
    n=3
    A = dom_eq_ran_mat
    Aop = odl.MatrixOperator(A)

    B = dom_eq_ran_mat+0.1
    Bop = odl.MatrixOperator(B)

    ABop = ProductSpaceOperator([[Aop],
                                 [Bop]])

    matrix_repr = matrix_representation(ABop)

    _, backend = get_array_and_backend(A)
    assert matrix_repr.shape == (2, n, 1, n)
    assert backend.to_cpu(
        backend.array_namespace.linalg.norm(A - matrix_repr[0, :, 0, :])) == pytest.approx(0)
    assert backend.to_cpu(
        backend.array_namespace.linalg.norm(B - matrix_repr[1, :, 0, :])) == pytest.approx(0)


def test_matrix_representation_product_to_product(dom_eq_ran_mat):
    """Verify that the matrix repr works for product spaces.

    Here, since the domain and range has shape ``(2, 3)``, the shape of the
    matrix representation will be ``(2, 3, 2, 3)``.
    """
    n = 3
    A = dom_eq_ran_mat
    Aop = odl.MatrixOperator(A)

    B = dom_eq_ran_mat+0.1
    Bop = odl.MatrixOperator(B)

    ABop = ProductSpaceOperator([[Aop, 0],
                                 [0, Bop]])
    matrix_repr = matrix_representation(ABop)

    assert matrix_repr.shape == (2, n, 2, n)
    _, backend = get_array_and_backend(A)
    assert matrix_repr.shape == (2, n, 2, n)
    assert backend.to_cpu(
        backend.array_namespace.linalg.norm(A - matrix_repr[0, :, 0, :])) == pytest.approx(0)
    assert backend.to_cpu(
        backend.array_namespace.linalg.norm(B - matrix_repr[1, :, 1, :])) == pytest.approx(0)



def test_matrix_representation_not_linear_op(odl_impl_device_pairs):
    """Verify error when operator is non-linear"""
    impl, device = odl_impl_device_pairs
    class MyNonLinOp(odl.Operator):
        """Small nonlinear test operator."""
        def _call(self, x):
            return x ** 2

    nonlin_op = MyNonLinOp(
        domain=odl.rn(3,impl=impl, device=device), 
        range=odl.rn(3,impl=impl, device=device), 
        linear=False)
    with pytest.raises(ValueError):
        matrix_representation(nonlin_op)


def test_matrix_representation_wrong_domain(odl_impl_device_pairs):
    """Verify that the matrix representation function gives correct error"""
    impl, device = odl_impl_device_pairs
    class MyOp(odl.Operator):
        """Small test operator."""
        def __init__(self):
            super(MyOp, self).__init__(
                domain=odl.rn(3,impl=impl, device=device) * odl.rn(3,impl=impl, device=device) ** 2,
                range=odl.rn(4,impl=impl, device=device),
                linear=True)

        def _call(self, x, out):
            return odl.rn([4], impl=impl, device=device)

    nonlin_op = MyOp()
    with pytest.raises(TypeError):
        matrix_representation(nonlin_op)


def test_matrix_representation_wrong_range(odl_impl_device_pairs):
    """Verify that the matrix representation function gives correct error"""
    impl, device = odl_impl_device_pairs
    class MyOp(odl.Operator):
        """Small test operator."""
        def __init__(self):
            super(MyOp, self).__init__(
                domain=odl.rn(3,impl=impl, device=device),
                range=odl.rn(3,impl=impl, device=device) * odl.rn(3,impl=impl, device=device) ** 2,
                linear=True)

        def _call(self, x, out):
             return odl.rn([4], impl=impl, device=device)

    nonlin_op = MyOp()
    with pytest.raises(TypeError):
        matrix_representation(nonlin_op)


def test_power_method_opnorm_symm(odl_impl_device_pairs):
    """Test the power method on a symmetrix matrix operator"""
    impl, device = odl_impl_device_pairs
    # Test matrix with singular values 1.2 and 1.0
    space = odl.rn([2,2], impl=impl, device=device)
    mat = space.element([[0.9509044, -0.64566614],
                    [-0.44583952, -0.95923051]])

    op = odl.MatrixOperator(mat)
    true_opnorm = 1.2

    # Run with a random starting point. This should give the right
    # result most of the time, but can fail if the random point fell
    # very close to the lesser eigenvector. To confirm that the probability
    # is low, we run the method three times (they will have different
    # random starting points). Usually all three should give the right
    # result, but one of them may occasionally fail (and give the other
    # singular value instead). Two wrong ones out of three should
    # be exeedingly rare though, so we treat this as a test failure.
    n_failures = 0
    for _ in range(3):
        opnorm_est = power_method_opnorm(op, maxiter=100)
        if n_failures<1:
            if opnorm_est != pytest.approx(true_opnorm, rel=1e-2):
                n_failures += 1
        else:
            assert opnorm_est == pytest.approx(true_opnorm, rel=1e-2)

    # Start at a deterministic point. This should _always_ succeed.
    xstart = odl.rn(2, impl=impl, device=device).element([0.8, 0.5])
    opnorm_est = power_method_opnorm(op, xstart=xstart, maxiter=100)
    assert opnorm_est == pytest.approx(true_opnorm, rel=1e-2)


def test_power_method_opnorm_nonsymm(odl_impl_device_pairs):
    """Test the power method on a nonsymmetrix matrix operator"""
    impl, device = odl_impl_device_pairs
    # Singular values 5.5 and 6
    space = odl.rn([3,2], impl=impl, device=device)
    mat = space.element([[-1.52441557, 5.04276365],
                    [1.90246927, 2.54424763],
                    [5.32935411, 0.04573162]])

    op = odl.MatrixOperator(mat)
    true_opnorm = 6

    # Start vector (1, 1) is close to the wrong eigenvector
    xstart = odl.rn(2, impl=impl, device=device).element([1, 1])
    opnorm_est = power_method_opnorm(op, xstart=xstart, maxiter=50)
    assert opnorm_est == pytest.approx(true_opnorm, rel=1e-2)

    # Start close to the correct eigenvector, converges very fast
    xstart = odl.rn(2, impl=impl, device=device).element([-0.8, 0.5])
    opnorm_est = power_method_opnorm(op, xstart=xstart, maxiter=6)
    assert opnorm_est == pytest.approx(true_opnorm, rel=1e-2)


def test_power_method_opnorm_exceptions(odl_impl_device_pairs):
    """Test the exceptions"""
    impl, device = odl_impl_device_pairs
    space = odl.rn(2, impl=impl, device=device)
    op = odl.IdentityOperator(space)

    with pytest.raises(ValueError):
        # Too small number of iterates
        power_method_opnorm(op, maxiter=0)

    with pytest.raises(ValueError):
        # Negative number of iterates
        power_method_opnorm(op, maxiter=-5)

    with pytest.raises(ValueError):
        # Input vector is zero
        power_method_opnorm(op, maxiter=2, xstart=space.zero())

    with pytest.raises(ValueError):
        # Input vector in the null space
        op = odl.MatrixOperator([[0., 1.],
                                 [0., 0.]])

        power_method_opnorm(op, maxiter=2, xstart=[1, 0])

    with pytest.raises(ValueError):
        # Uneven number of iterates for non square operator
        op = odl.MatrixOperator([[1., 2., 3.],
                                 [4., 5., 6.]])

        power_method_opnorm(op, maxiter=1, xstart=op.domain.one())


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import pytest
import odl


def test_all_imports():
    # Create Cn
    odl.cn(3)
    odl.core.space.cn(3)
    C3 = odl.core.space.space_utils.cn(3)

    # Three ways of creating the identity
    odl.IdentityOperator(C3)
    odl.core.operator.IdentityOperator(C3)
    odl.core.operator.default_ops.IdentityOperator(C3)

    # Test that utility needs to be explicitly imported
    odl.core.util.print_utils.array_str
    with pytest.raises(AttributeError):
        odl.array_str


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import odl
import numpy as np

from odl.core.util.dtype_utils import (
    is_numeric_dtype, is_real_dtype, is_floating_dtype,
    is_complex_dtype,
    FLOAT_DTYPES,
    COMPLEX_DTYPES,
    INTEGER_DTYPES    
    )


real_float_dtypes = FLOAT_DTYPES
complex_float_dtypes = COMPLEX_DTYPES
nonfloat_numeric_dtypes = INTEGER_DTYPES
numeric_dtypes = (real_float_dtypes + complex_float_dtypes +
                  nonfloat_numeric_dtypes)
real_dtypes = real_float_dtypes + nonfloat_numeric_dtypes
# Need to make concrete instances here (with string lengths)

# ---- Data type helpers ---- #


def test_is_numeric_dtype():
    for dtype in numeric_dtypes:
        assert is_numeric_dtype(dtype)


def test_is_real_dtype():
    for dtype in real_dtypes:
        assert is_real_dtype(dtype)


def test_is_real_floating_dtype():
    for dtype in real_float_dtypes:
        assert is_floating_dtype(dtype)


def test_is_complex_floating_dtype():
    for dtype in complex_float_dtypes:
        assert is_complex_dtype(dtype)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Satistical functions expected by the python array API
Internally, all functions apply a reduction operation on a LinearSpaceElement.

Args:
    x (LinearSpaceElement): LinearSpaceElement on which to apply the reduction. 

Returns:
    x (float | array-like): Output of the reduction. 

Notes:
    1) The actual implementation of the reduction is in the LinearSpace of this element.
    2) These functions can return python Numbers or backend-specific array (when calling with keepdims=True for instance), but they will not return odl objects.

"""

__all__ = (
    'cumulative_prod',
    'cumulative_sum',
    'max',
    'mean',
    'min',
    'prod',
    'std',
    'sum',
    'var'   
)

def _apply_reduction(operation: str, x, **kwargs):
    """
    Examples
    >>> e1 = odl.rn(3).element((1,2,3))
    >>> odl.cumulative_prod(e1) == [1,2,6]
    array([ True,  True,  True], dtype=bool)
    >>> odl.cumulative_sum(e1) == [1,3,6]
    array([ True,  True,  True], dtype=bool)
    >>> odl.max(e1) == 3
    True
    >>> odl.mean(e1) == 2
    True
    >>> odl.min(e1) == 1
    True
    >>> odl.prod(e1) == 6
    True
    >>> odl.std(e1) == np.std([1,2,3])
    True
    >>> odl.sum(e1) == 6
    True
    >>> odl.var(e1) == np.var([1,2,3])
    True
    """
    return x.space._element_reduction(operation=operation, x=x, **kwargs)

def cumulative_prod(x, axis=None, dtype=None, include_initial=False):
    """
    Calculates the cumulative product of elements in the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('cumulative_prod', x, axis=axis, dtype=dtype, include_initial=include_initial)

def cumulative_sum(x, axis=None, dtype=None, include_initial=False):
    """
    Calculates the cumulative sum of elements in the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('cumulative_sum', x, axis=axis, dtype=dtype, include_initial=include_initial)

def max(x, axis=None, keepdims=False):
    """
    Calculates the maximum value of the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('max', x, axis=axis, keepdims=keepdims)

def mean(x, axis=None, keepdims=False):
    """
    Calculates the arithmetic mean of the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('mean', x, axis=axis, keepdims=keepdims)

def min(x, axis=None, keepdims=False):
    """
    Calculates the minimum value of the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('min', x, axis=axis, keepdims=keepdims)

def prod(x, axis=None, dtype=None, keepdims=False):
    """
    Calculates the product of input array x elements.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('prod', x, axis=axis, dtype=dtype, keepdims=keepdims)

def std(x, axis=None, correction=0.0, keepdims=False):
    """
    Calculates the standard deviation of the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('std', x, axis=axis, correction=correction, keepdims=keepdims)

def sum(x, axis=None, dtype=None, keepdims=False):
    """
    Calculates the sum of the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('sum', x, axis=axis, dtype=dtype, keepdims=keepdims)

def var(x, axis=None, correction=0.0, keepdims=False):
    """    
    Calculates the variance of the input array x.
    Note: This function might not be doing what you expect. If you want to return an array (np.ndarray, torch.Tensor...), you are in the right place. However, you cannot use it to create a new LinearSpaceSelement.
    """
    return _apply_reduction('var', x, axis=axis, correction=correction, keepdims=keepdims)
